# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for DebianSuiteManager."""

import datetime as dt
import re
from pathlib import PurePath
from typing import ClassVar, override

from django.contrib.auth import get_user_model
from django.utils import timezone
from pgpy import PGPKey, PGPKeyring
from pgpy.constants import EllipticCurveOID, PubKeyAlgorithm

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.assets.models import KeyPurpose
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    FileInArtifact,
    User,
    Workspace,
    default_workspace,
)
from debusine.db.permissioncontext import PermissionContext
from debusine.server.collections import DebianSuiteManager, ItemAdditionError
from debusine.server.collections.debian_suite import (
    make_pool_filename,
    make_source_prefix,
)
from debusine.test.django import TestCase


class MakePoolFilenameTests(TestCase):
    """Tests for make_pool_filename."""

    def test_normal(self) -> None:
        """`make_pool_filename` works for normal source package names."""
        self.assertEqual(make_source_prefix("hello"), "h")
        self.assertEqual(
            "pool/main/h/hello/hello_1.0.dsc",
            make_pool_filename("hello", "main", "hello_1.0.dsc"),
        )

    def test_library(self) -> None:
        """`make_pool_filename` works for library source package names."""
        self.assertEqual(make_source_prefix("libhello"), "libh")
        self.assertEqual(
            "pool/non-free/libh/libhello/libhello_1.0.dsc",
            make_pool_filename("libhello", "non-free", "libhello_1.0.dsc"),
        )


class DebianSuiteManagerTests(TestCase):
    """Tests for DebianSuiteManager."""

    user: ClassVar[User]
    workspace: ClassVar[Workspace]
    collection: ClassVar[Collection]
    manager: ClassVar[DebianSuiteManager]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()

        cls.user = get_user_model().objects.create_user(
            username="John", email="john@example.org"
        )

        cls.workspace = default_workspace()

        cls.collection = Collection.objects.create(
            name="Debian",
            category=CollectionCategory.SUITE,
            workspace=cls.workspace,
        )

        assert isinstance(cls.collection.manager, DebianSuiteManager)
        cls.manager = cls.collection.manager

    def create_source_package(
        self, name: str, version: str, paths: list[str]
    ) -> Artifact:
        """Create a minimal `debian:source-package` artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": name,
                "version": version,
                "type": "dpkg",
                "dsc_fields": {},
            },
            paths=paths,
            create_files=True,
            skip_add_files_in_store=True,
        )
        return artifact

    def create_binary_package(
        self,
        srcpkg_name: str,
        srcpkg_version: str,
        name: str,
        version: str,
        architecture: str,
        paths: list[str],
    ) -> Artifact:
        """Create a minimal `debian:binary-package` artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": srcpkg_name,
                "srcpkg_version": srcpkg_version,
                "deb_fields": {
                    "Package": name,
                    "Version": version,
                    "Architecture": architecture,
                },
                "deb_control_files": [],
            },
            paths=paths,
            create_files=True,
            skip_add_files_in_store=True,
        )
        return artifact

    def test_init_wrong_collection_category_raise_value_error(self) -> None:
        """Init raise ValueError: wrong collection category."""
        category = "debian:something-else"
        collection = Collection.objects.create(
            name="Name is not used",
            category=category,
            workspace=self.workspace,
        )

        msg = f'^DebianSuiteManager cannot manage "{category}" category$'

        with self.assertRaisesRegex(ValueError, msg):
            DebianSuiteManager(collection)

    def test_parent_archive(self) -> None:
        self.assertEqual(
            self.manager.parent_archive,
            self.workspace.collections.get(category=CollectionCategory.ARCHIVE),
        )

    def test_parent_archive_none(self) -> None:
        self.workspace.collections.filter(
            category=CollectionCategory.ARCHIVE
        ).delete()

        self.assertIsNone(self.manager.parent_archive)

    def test_may_reuse_versions_suite_reuses_versions(self) -> None:
        archive = self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE
        )
        self.collection.data["may_reuse_versions"] = True
        self.collection.save()
        archive.manager.add_collection(self.collection, user=self.user)

        self.assertTrue(self.manager.may_reuse_versions)

    def test_may_reuse_versions_no_parent_archive(self) -> None:
        self.workspace.collections.filter(
            category=CollectionCategory.ARCHIVE
        ).delete()

        self.assertFalse(self.manager.may_reuse_versions)

    def test_may_reuse_versions_archive_reuses_versions(self) -> None:
        archive = self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE
        )
        archive.data["may_reuse_versions"] = True
        archive.save()
        archive.manager.add_collection(self.collection, user=self.user)

        self.assertTrue(self.manager.may_reuse_versions)

    def test_may_reuse_versions_neither_archive_nor_suite_reuses_versions(
        self,
    ) -> None:
        archive = self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE
        )
        archive.manager.add_collection(self.collection, user=self.user)

        self.assertFalse(self.manager.may_reuse_versions)

    def test_get_signing_keys_from_suite(self) -> None:
        """Signing keys may be configured directly on the suite."""
        self.assertQuerySetEqual(self.manager.get_signing_keys(), [])

        suite_fingerprint = "0" * 40
        signing_key_asset = self.playground.create_signing_key_asset(
            purpose=KeyPurpose.OPENPGP_REPOSITORY,
            fingerprint=suite_fingerprint,
            workspace=self.workspace,
        )
        self.playground.create_asset_usage(
            resource=signing_key_asset, workspace=self.workspace
        )
        self.collection.data["signing_keys"] = [suite_fingerprint]
        self.collection.save()

        self.assertQuerySetEqual(
            self.manager.get_signing_keys(), [signing_key_asset]
        )

    def test_get_signing_keys_from_parent_archive(self) -> None:
        """Signing keys may be configured on the parent archive."""
        archive = self.workspace.get_singleton_collection(
            pc=PermissionContext.for_ui_user(self.user),
            category=CollectionCategory.ARCHIVE,
        )
        archive.manager.add_collection(self.collection, user=self.user)

        self.assertQuerySetEqual(self.manager.get_signing_keys(), [])

        archive_fingerprint = "1" * 40
        signing_key_asset = self.playground.create_signing_key_asset(
            purpose=KeyPurpose.OPENPGP_REPOSITORY,
            fingerprint=archive_fingerprint,
            workspace=self.workspace,
        )
        self.playground.create_asset_usage(
            resource=signing_key_asset, workspace=self.workspace
        )
        archive.data["signing_keys"] = [archive_fingerprint]
        archive.save()

        self.assertQuerySetEqual(
            self.manager.get_signing_keys(), [signing_key_asset]
        )

    def test_get_signing_keys_no_parent_archive(self) -> None:
        """``get_signing_keys`` returns None if there is no parent archive."""
        self.workspace.collections.filter(
            category=CollectionCategory.ARCHIVE
        ).delete()

        self.assertQuerySetEqual(self.manager.get_signing_keys(), [])

    def test_get_signing_keys_no_asset_usage(self) -> None:
        """``get_signing_keys`` only returns keys used in this workspace."""
        suite_fingerprint = "0" * 40
        self.playground.create_signing_key_asset(
            purpose=KeyPurpose.OPENPGP_REPOSITORY,
            fingerprint=suite_fingerprint,
            workspace=self.workspace,
        )
        self.collection.data["signing_keys"] = [suite_fingerprint]
        self.collection.save()

        self.assertQuerySetEqual(self.manager.get_signing_keys(), [])

    def test_export_signing_keys_none(self) -> None:
        self.assertIsNone(self.manager.export_signing_keys())

    def test_export_signing_keys(self) -> None:
        keys = [
            PGPKey.new(PubKeyAlgorithm.EdDSA, EllipticCurveOID.Ed25519).pubkey
            for _ in range(2)
        ]
        for key in keys:
            self.playground.create_asset_usage(
                resource=self.playground.create_signing_key_asset(
                    purpose=KeyPurpose.OPENPGP_REPOSITORY,
                    fingerprint=key.fingerprint,
                    public_key=str(key),
                    workspace=self.workspace,
                ),
                workspace=self.workspace,
            )
        self.collection.data["signing_keys"] = [key.fingerprint for key in keys]
        self.collection.save()

        self.assertEqual(
            PGPKeyring(self.manager.export_signing_keys()).fingerprints(),
            {key.fingerprint for key in keys},
        )

    def test_add_artifact_source_package(self) -> None:
        """`add_artifact` can add a source package to the suite."""
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        dsc_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.dsc"
        ).file
        tar_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.tar.xz"
        ).file

        collection_item = self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.name, "hello_1.0")
        self.assertEqual(
            collection_item.data,
            {
                "package": "hello",
                "version": "1.0",
                "component": "main",
                "section": "devel",
            },
        )
        self.assert_collection_has_constraints(
            self.collection,
            [
                {
                    "collection_item_id": collection_item.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/hello_1.0.dsc",
                    "value": f"sha256:{dsc_file.hash_digest.hex()}",
                },
                {
                    "collection_item_id": collection_item.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/hello_1.0.tar.xz",
                    "value": f"sha256:{tar_file.hash_digest.hex()}",
                },
            ],
        )

    def test_add_artifact_source_package_adds_archive_constraints(self) -> None:
        """`add_artifact` adds archive constraints for source packages."""
        archive = self.workspace.get_singleton_collection(
            pc=PermissionContext.for_ui_user(self.user),
            category=CollectionCategory.ARCHIVE,
        )
        archive.manager.add_collection(self.collection, user=self.user)
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        dsc_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.dsc"
        ).file
        tar_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.tar.xz"
        ).file

        collection_item = self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.name, "hello_1.0")
        self.assertEqual(
            collection_item.data,
            {
                "package": "hello",
                "version": "1.0",
                "component": "main",
                "section": "devel",
            },
        )
        for collection in (self.collection, archive):
            self.assert_collection_has_constraints(
                collection,
                [
                    {
                        "collection_item_id": collection_item.id,
                        "constraint_name": "pool-file",
                        "key": "pool/main/h/hello/hello_1.0.dsc",
                        "value": f"sha256:{dsc_file.hash_digest.hex()}",
                    },
                    {
                        "collection_item_id": collection_item.id,
                        "constraint_name": "pool-file",
                        "key": "pool/main/h/hello/hello_1.0.tar.xz",
                        "value": f"sha256:{tar_file.hash_digest.hex()}",
                    },
                ],
            )

    def test_add_artifact_source_package_replace(self) -> None:
        """`add_artifact` can replace an existing source package."""
        self.collection.data["may_reuse_versions"] = True
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        collection_item = self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )
        source_package_artifact2 = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        collection_item2 = self.manager.add_artifact(
            source_package_artifact2,
            user=self.user,
            variables={"component": "main", "section": "devel"},
            replace=True,
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.artifact, source_package_artifact)
        self.assertEqual(collection_item.removed_by_user, self.user)
        self.assertIsNotNone(collection_item.removed_at)
        self.assertEqual(collection_item2.name, "hello_1.0")
        self.assertEqual(collection_item2.artifact, source_package_artifact2)
        self.assertIsNone(collection_item2.removed_at)

    def test_add_artifact_source_package_replace_nonexistent(self) -> None:
        """Replacing a nonexistent source package is allowed."""
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )

        collection_item = self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
            replace=True,
        )

        self.assertEqual(collection_item.name, "hello_1.0")
        self.assertEqual(collection_item.artifact, source_package_artifact)

    def test_add_artifact_source_package_already_replaced(self) -> None:
        """`add_artifact` can add an already-replaced source package."""
        self.collection.data["may_reuse_versions"] = True
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        dsc_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.dsc"
        ).file
        tar_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.tar.xz"
        ).file
        collection_item_new = self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )
        source_package_artifact_old = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.gz"]
        )
        yesterday = timezone.now() - dt.timedelta(days=1)
        collection_item_old = self.manager.add_artifact(
            source_package_artifact_old,
            user=self.user,
            variables={"component": "main", "section": "devel"},
            created_at=yesterday,
            replaced_by=collection_item_new,
        )

        self.assertEqual(collection_item_new.artifact, source_package_artifact)
        self.assertIsNone(collection_item_new.removed_at)
        self.assertEqual(collection_item_old.name, "hello_1.0")
        self.assertEqual(
            collection_item_old.artifact, source_package_artifact_old
        )
        self.assertEqual(collection_item_old.created_at, yesterday)
        self.assertEqual(
            collection_item_old.removed_at, collection_item_new.created_at
        )
        self.assertEqual(
            collection_item_old.removed_by_user,
            collection_item_new.created_by_user,
        )
        self.assert_collection_has_constraints(
            self.collection,
            [
                {
                    "collection_item_id": collection_item_new.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/hello_1.0.dsc",
                    "value": f"sha256:{dsc_file.hash_digest.hex()}",
                },
                {
                    "collection_item_id": collection_item_new.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/hello_1.0.tar.xz",
                    "value": f"sha256:{tar_file.hash_digest.hex()}",
                },
            ],
        )

    def test_add_artifact_source_package_idempotent(self) -> None:
        """`add_artifact` can add the same source package twice."""
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        collection_item = self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )

        self.assertEqual(
            self.manager.add_artifact(
                source_package_artifact,
                user=self.user,
                variables={"component": "main", "section": "devel"},
            ),
            collection_item,
        )

    def test_add_artifact_source_package_older_version(self) -> None:
        """Suites normally forbid rewinding source packages."""
        source_package_artifact = self.create_source_package(
            "hello", "1.1", ["hello_1.1.dsc", "hello_1.1.tar.xz"]
        )
        self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )
        source_package_artifact2 = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        with self.assertRaisesRegex(
            ItemAdditionError,
            re.escape(
                f"{self.collection}: New source version hello 1.0"
                f" < existing version 1.1"
            ),
        ):
            self.manager.add_artifact(
                source_package_artifact2,
                user=self.user,
                variables={"component": "main", "section": "devel"},
                replace=True,
            )

        # This can be overridden temporarily.
        self.assertFalse(self.manager.may_rewind_versions)
        with self.manager.allow_rewinding_versions():
            self.assertTrue(self.manager.may_rewind_versions)
            self.manager.add_artifact(
                source_package_artifact2,
                user=self.user,
                variables={"component": "main", "section": "devel"},
                replace=True,
            )
        self.assertFalse(self.manager.may_rewind_versions)

    def test_add_artifact_source_package_may_rewind_versions(self) -> None:
        """Suites can be configured to allow rewinding source packages."""
        self.collection.data["may_rewind_versions"] = True
        source_package_artifact = self.create_source_package(
            "hello", "1.1", ["hello_1.1.dsc", "hello_1.1.tar.xz"]
        )
        collection_item = self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )
        source_package_artifact2 = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        collection_item2 = self.manager.add_artifact(
            source_package_artifact2,
            user=self.user,
            variables={"component": "main", "section": "devel"},
            replace=True,
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.artifact, source_package_artifact)
        self.assertIsNone(collection_item.removed_by_user)
        self.assertIsNone(collection_item.removed_at)
        self.assertEqual(collection_item2.artifact, source_package_artifact2)
        self.assertIsNone(collection_item2.removed_by_user)
        self.assertIsNone(collection_item2.removed_at)

    def test_add_artifact_binary_package(self) -> None:
        """`add_artifact` can add a binary package to the suite."""
        binary_package_artifact = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb", "libhello1-dbgsym_1.0_amd64.deb"],
        )
        deb_file = binary_package_artifact.fileinartifact_set.get(
            path="libhello1_1.0_amd64.deb"
        ).file
        dbgsym_file = binary_package_artifact.fileinartifact_set.get(
            path="libhello1-dbgsym_1.0_amd64.deb",
        ).file

        collection_item = self.manager.add_artifact(
            binary_package_artifact,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.name, "libhello1_1:1.0_amd64")
        self.assertEqual(
            collection_item.data,
            {
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0",
                "package": "libhello1",
                "version": "1:1.0",
                "architecture": "amd64",
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )
        self.assert_collection_has_constraints(
            self.collection,
            [
                {
                    "collection_item_id": collection_item.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/libhello1_1.0_amd64.deb",
                    "value": f"sha256:{deb_file.hash_digest.hex()}",
                },
                {
                    "collection_item_id": collection_item.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/libhello1-dbgsym_1.0_amd64.deb",
                    "value": f"sha256:{dbgsym_file.hash_digest.hex()}",
                },
            ],
        )

    def test_add_artifact_binary_package_adds_archive_constraints(self) -> None:
        """`add_artifact` adds archive constraints for binary packages."""
        archive = self.workspace.get_singleton_collection(
            pc=PermissionContext.for_ui_user(self.user),
            category=CollectionCategory.ARCHIVE,
        )
        archive.manager.add_collection(self.collection, user=self.user)
        binary_package_artifact = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb", "libhello1-dbgsym_1.0_amd64.deb"],
        )
        deb_file = binary_package_artifact.fileinartifact_set.get(
            path="libhello1_1.0_amd64.deb"
        ).file
        dbgsym_file = binary_package_artifact.fileinartifact_set.get(
            path="libhello1-dbgsym_1.0_amd64.deb",
        ).file

        collection_item = self.manager.add_artifact(
            binary_package_artifact,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.name, "libhello1_1:1.0_amd64")
        self.assertEqual(
            collection_item.data,
            {
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0",
                "package": "libhello1",
                "version": "1:1.0",
                "architecture": "amd64",
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )
        for collection in (self.collection, archive):
            self.assert_collection_has_constraints(
                collection,
                [
                    {
                        "collection_item_id": collection_item.id,
                        "constraint_name": "pool-file",
                        "key": "pool/main/h/hello/libhello1_1.0_amd64.deb",
                        "value": f"sha256:{deb_file.hash_digest.hex()}",
                    },
                    {
                        "collection_item_id": collection_item.id,
                        "constraint_name": "pool-file",
                        "key": (
                            "pool/main/h/hello/libhello1-dbgsym_1.0_amd64.deb"
                        ),
                        "value": f"sha256:{dbgsym_file.hash_digest.hex()}",
                    },
                ],
            )

    def test_add_artifact_binary_package_replace(self) -> None:
        """`add_artifact` can replace an existing binary package."""
        self.collection.data["may_reuse_versions"] = True
        binary_package_artifact = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb", "libhello1-dbgsym_1.0_amd64.deb"],
        )
        collection_item = self.manager.add_artifact(
            binary_package_artifact,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )
        binary_package_artifact2 = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb", "libhello1-dbgsym_1.0_amd64.deb"],
        )
        collection_item2 = self.manager.add_artifact(
            binary_package_artifact2,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
            replace=True,
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.artifact, binary_package_artifact)
        self.assertEqual(collection_item.removed_by_user, self.user)
        self.assertIsNotNone(collection_item.removed_at)
        self.assertEqual(collection_item2.name, "libhello1_1:1.0_amd64")
        self.assertEqual(collection_item2.artifact, binary_package_artifact2)
        self.assertIsNone(collection_item2.removed_at)

    def test_add_artifact_binary_package_replace_nonexistent(self) -> None:
        """Replacing a nonexistent binary package is allowed."""
        binary_package_artifact = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb", "libhello1-dbgsym_1.0_amd64.deb"],
        )

        collection_item = self.manager.add_artifact(
            binary_package_artifact,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
            replace=True,
        )

        self.assertEqual(collection_item.name, "libhello1_1:1.0_amd64")
        self.assertEqual(collection_item.artifact, binary_package_artifact)

    def test_add_artifact_binary_package_already_replaced(self) -> None:
        """`add_artifact` can add an already-replaced binary package."""
        self.collection.data["may_reuse_versions"] = True
        binary_package_artifact = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb", "libhello1-dbgsym_1.0_amd64.deb"],
        )
        deb_file = binary_package_artifact.fileinartifact_set.get(
            path="libhello1_1.0_amd64.deb"
        ).file
        dbgsym_file = binary_package_artifact.fileinartifact_set.get(
            path="libhello1-dbgsym_1.0_amd64.deb",
        ).file
        collection_item_new = self.manager.add_artifact(
            binary_package_artifact,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )
        binary_package_artifact_old = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb", "libhello-dev_1.0_amd64.deb"],
        )
        yesterday = timezone.now() - dt.timedelta(days=1)
        collection_item_old = self.manager.add_artifact(
            binary_package_artifact_old,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
            created_at=yesterday,
            replaced_by=collection_item_new,
        )

        self.assertEqual(collection_item_new.artifact, binary_package_artifact)
        self.assertIsNone(collection_item_new.removed_at)
        self.assertEqual(collection_item_old.name, "libhello1_1:1.0_amd64")
        self.assertEqual(
            collection_item_old.artifact, binary_package_artifact_old
        )
        self.assertEqual(collection_item_old.created_at, yesterday)
        self.assertEqual(
            collection_item_old.removed_at, collection_item_new.created_at
        )
        self.assertEqual(
            collection_item_old.removed_by_user,
            collection_item_new.created_by_user,
        )
        self.assert_collection_has_constraints(
            self.collection,
            [
                {
                    "collection_item_id": collection_item_new.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/libhello1_1.0_amd64.deb",
                    "value": f"sha256:{deb_file.hash_digest.hex()}",
                },
                {
                    "collection_item_id": collection_item_new.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/libhello1-dbgsym_1.0_amd64.deb",
                    "value": f"sha256:{dbgsym_file.hash_digest.hex()}",
                },
            ],
        )

    def test_add_artifact_binary_package_idempotent(self) -> None:
        """`add_artifact` can add the same binary package twice."""
        binary_package_artifact = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb", "libhello1-dbgsym_1.0_amd64.deb"],
        )
        collection_item = self.manager.add_artifact(
            binary_package_artifact,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        self.assertEqual(
            self.manager.add_artifact(
                binary_package_artifact,
                user=self.user,
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
            ),
            collection_item,
        )

    def test_add_artifact_binary_package_older_version(self) -> None:
        """Suites normally forbid rewinding binary packages."""
        binary_package_artifact = self.create_binary_package(
            "hello",
            "1.1",
            "libhello1",
            "1:1.1",
            "amd64",
            ["libhello1_1.1_amd64.deb"],
        )
        self.manager.add_artifact(
            binary_package_artifact,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )
        binary_package_artifact2 = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb"],
        )
        with self.assertRaisesRegex(
            ItemAdditionError,
            re.escape(
                f"{self.collection}: New binary version libhello1:amd64 1:1.0"
                f" < existing version 1:1.1"
            ),
        ):
            self.manager.add_artifact(
                binary_package_artifact2,
                user=self.user,
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
                replace=True,
            )

        # This can be overridden temporarily.
        self.assertFalse(self.manager.may_rewind_versions)
        with self.manager.allow_rewinding_versions():
            self.assertTrue(self.manager.may_rewind_versions)
            self.manager.add_artifact(
                binary_package_artifact2,
                user=self.user,
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
                replace=True,
            )
        self.assertFalse(self.manager.may_rewind_versions)

    def test_add_artifact_binary_package_may_rewind_versions(self) -> None:
        """Suites can be configured to allow rewinding binary packages."""
        self.collection.data["may_rewind_versions"] = True
        binary_package_artifact = self.create_binary_package(
            "hello",
            "1.1",
            "libhello1",
            "1:1.1",
            "amd64",
            ["libhello1_1.1_amd64.deb"],
        )
        collection_item = self.manager.add_artifact(
            binary_package_artifact,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )
        binary_package_artifact2 = self.create_binary_package(
            "hello",
            "1.0",
            "libhello1",
            "1:1.0",
            "amd64",
            ["libhello1_1.0_amd64.deb"],
        )
        collection_item2 = self.manager.add_artifact(
            binary_package_artifact2,
            user=self.user,
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
            replace=True,
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.artifact, binary_package_artifact)
        self.assertIsNone(collection_item.removed_by_user)
        self.assertIsNone(collection_item.removed_at)
        self.assertEqual(collection_item2.artifact, binary_package_artifact2)
        self.assertIsNone(collection_item2.removed_by_user)
        self.assertIsNone(collection_item2.removed_at)

    def test_add_artifact_repository_index(self) -> None:
        """``add_artifact`` can add a repository index to the suite."""
        packages_artifact = self.playground.create_repository_index(
            "Packages.xz", skip_add_files_in_store=True
        )

        collection_item = self.manager.add_artifact(
            packages_artifact,
            user=self.user,
            variables={"path": "main/binary-amd64/Packages.xz"},
        )

        self.assertEqual(
            collection_item.name, "index:main/binary-amd64/Packages.xz"
        )
        self.assertEqual(
            collection_item.data, {"path": "main/binary-amd64/Packages.xz"}
        )
        self.assert_collection_has_constraints(self.collection, [])

    def test_add_artifact_repository_index_replace(self) -> None:
        """``add_artifact`` can replace an existing repository index."""
        packages_artifact = self.playground.create_repository_index(
            "Packages.xz", skip_add_files_in_store=True
        )
        collection_item = self.manager.add_artifact(
            packages_artifact,
            user=self.user,
            variables={"path": "main/binary-amd64/Packages.xz"},
        )
        packages_artifact2 = self.playground.create_repository_index(
            "Packages.xz", skip_add_files_in_store=True
        )
        collection_item2 = self.manager.add_artifact(
            packages_artifact2,
            user=self.user,
            variables={"path": "main/binary-amd64/Packages.xz"},
            replace=True,
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.artifact, packages_artifact)
        self.assertEqual(collection_item.removed_by_user, self.user)
        self.assertIsNotNone(collection_item.removed_at)
        self.assertEqual(
            collection_item2.name, "index:main/binary-amd64/Packages.xz"
        )
        self.assertEqual(collection_item2.artifact, packages_artifact2)
        self.assertIsNone(collection_item2.removed_at)

    def test_add_artifact_repository_index_already_replaced(self) -> None:
        """``add_artifact`` can add an already-replaced repository index."""
        packages_artifact = self.playground.create_repository_index(
            "Packages.xz", skip_add_files_in_store=True
        )
        collection_item_new = self.manager.add_artifact(
            packages_artifact,
            user=self.user,
            variables={"path": "main/binary-amd64/Packages.xz"},
        )
        packages_artifact2 = self.playground.create_repository_index(
            "Packages.xz", skip_add_files_in_store=True
        )
        yesterday = timezone.now() - dt.timedelta(days=1)
        collection_item_old = self.manager.add_artifact(
            packages_artifact2,
            user=self.user,
            variables={"path": "main/binary-amd64/Packages.xz"},
            created_at=yesterday,
            replaced_by=collection_item_new,
        )

        self.assertEqual(collection_item_new.artifact, packages_artifact)
        self.assertIsNone(collection_item_new.removed_at)
        self.assertEqual(
            collection_item_old.name, "index:main/binary-amd64/Packages.xz"
        )
        self.assertEqual(collection_item_old.artifact, packages_artifact2)
        self.assertEqual(collection_item_old.created_at, yesterday)
        self.assertEqual(
            collection_item_old.removed_at, collection_item_new.created_at
        )
        self.assertEqual(
            collection_item_old.removed_by_user,
            collection_item_new.created_by_user,
        )

    def test_add_artifact_repository_index_idempotent(self) -> None:
        """``add_artifact`` can add the same repository index twice."""
        packages_artifact = self.playground.create_repository_index(
            "Packages.xz", skip_add_files_in_store=True
        )
        collection_item = self.manager.add_artifact(
            packages_artifact,
            user=self.user,
            variables={"path": "main/binary-amd64/Packages.xz"},
        )

        self.assertEqual(
            self.manager.add_artifact(
                packages_artifact,
                user=self.user,
                variables={"path": "main/binary-amd64/Packages.xz"},
            ),
            collection_item,
        )

    def test_do_add_artifact_source_package_requires_component(self) -> None:
        """`do_add_artifact` requires `component` for source packages."""
        source_package_artifact = self.create_source_package("hello", "1.0", [])

        with self.assertRaisesRegex(
            ItemAdditionError,
            '^Adding debian:source-package to debian:suite requires '
            '"component"$',
        ):
            self.manager.do_add_artifact(
                source_package_artifact, user=self.user
            )

    def test_do_add_artifact_source_package_requires_section(self) -> None:
        """`do_add_artifact` requires `section` for source packages."""
        source_package_artifact = self.create_source_package("hello", "1.0", [])

        with self.assertRaisesRegex(
            ItemAdditionError,
            '^Adding debian:source-package to debian:suite requires "section"$',
        ):
            self.manager.do_add_artifact(
                source_package_artifact,
                user=self.user,
                variables={"component": "main"},
            )

    def test_do_add_artifact_binary_package_requires_component(self) -> None:
        """`do_add_artifact` requires `component` for binary packages."""
        binary_package_artifact = self.create_binary_package(
            "hello", "1.0", "hello", "1.0", "amd64", []
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            '^Adding debian:binary-package to debian:suite requires '
            '"component"$',
        ):
            self.manager.do_add_artifact(
                binary_package_artifact, user=self.user
            )

    def test_do_add_artifact_binary_package_requires_section(self) -> None:
        """`do_add_artifact` requires `section` for binary packages."""
        binary_package_artifact = self.create_binary_package(
            "hello", "1.0", "hello", "1.0", "amd64", []
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            '^Adding debian:binary-package to debian:suite requires "section"$',
        ):
            self.manager.do_add_artifact(
                binary_package_artifact,
                user=self.user,
                variables={"component": "main"},
            )

    def test_do_add_artifact_binary_package_requires_priority(self) -> None:
        """`do_add_artifact` requires `priority` for binary packages."""
        binary_package_artifact = self.create_binary_package(
            "hello", "1.0", "hello", "1.0", "amd64", []
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            '^Adding debian:binary-package to debian:suite requires '
            '"priority"$',
        ):
            self.manager.do_add_artifact(
                binary_package_artifact,
                user=self.user,
                variables={"component": "main", "section": "devel"},
            )

    def test_do_add_artifact_overlapping_files_same_contents(self) -> None:
        """`do_add_artifact` allows same paths with same contents."""
        source_package_artifacts: list[Artifact] = []
        path_digests: list[tuple[str, str]] = []
        for version in ("1.0-1", "1.0-2"):
            artifact = self.create_source_package("hello", version, [])
            for path in (
                f"hello_{version}.dsc",
                f"hello_{version}.debian.tar.xz",
                "hello_1.0.orig.tar.xz",
            ):
                contents = f"data for {path}".encode()
                file = self.playground.create_file(contents)
                FileInArtifact.objects.create(
                    artifact=artifact, path=path, file=file
                )
                path_digests.append((path, file.hash_digest.hex()))
            source_package_artifacts.append(artifact)

        collection_items = [
            self.manager.add_artifact(
                artifact,
                user=self.user,
                variables={"component": "main", "section": "devel"},
            )
            for artifact in source_package_artifacts
        ]

        for collection_item in collection_items:
            collection_item.refresh_from_db()
        self.assertEqual(
            [item.name for item in collection_items],
            ["hello_1.0-1", "hello_1.0-2"],
        )
        self.assertEqual(
            [item.data for item in collection_items],
            [
                {
                    "package": "hello",
                    "version": version,
                    "component": "main",
                    "section": "devel",
                }
                for version in ("1.0-1", "1.0-2")
            ],
        )
        self.assert_collection_has_constraints(
            self.collection,
            [
                {
                    "collection_item_id": collection_items[item_index].id,
                    "constraint_name": "pool-file",
                    "key": f"pool/main/h/hello/{path}",
                    "value": f"sha256:{digest}",
                }
                for item_index, (path, digest) in zip(
                    (0, 0, 0, 1, 1, 1), path_digests
                )
            ],
        )

    def test_do_add_artifact_overlapping_files_different_contents(self) -> None:
        """`do_add_artifact` forbids same paths with different contents."""
        source_package_artifacts: list[Artifact] = []
        for version in ("1.0-1", "1.0-2"):
            artifact = self.create_source_package("hello", version, [])
            for path in (
                f"hello_{version}.dsc",
                f"hello_{version}.debian.tar.xz",
                "hello_1.0.orig.tar.xz",
            ):
                contents = f"data for {path} at {version}".encode()
                file = self.playground.create_file(contents)
                FileInArtifact.objects.create(
                    artifact=artifact, path=path, file=file
                )
            source_package_artifacts.append(artifact)

        self.manager.add_artifact(
            source_package_artifacts[0],
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )
        with self.assertRaisesRegex(
            ItemAdditionError, "db_collectionitemmatchconstraint_match_value"
        ):
            self.manager.add_artifact(
                source_package_artifacts[1],
                user=self.user,
                variables={"component": "main", "section": "devel"},
            )

    def test_do_remove_item_artifact(self) -> None:
        """`do_remove_item` removes an artifact item."""
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )
        dsc_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.dsc"
        ).file
        tar_file = source_package_artifact.fileinartifact_set.get(
            path="hello_1.0.tar.xz"
        ).file

        collection_item = self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )
        self.manager.remove_item(collection_item, user=self.user)

        collection_item.refresh_from_db()

        # The artifact is not removed yet (retention period applies)
        self.assertEqual(collection_item.artifact, source_package_artifact)

        self.assertEqual(collection_item.removed_by_user, self.user)
        self.assertIsInstance(collection_item.removed_at, dt.datetime)

        # Since `may_reuse_versions` defaults to False, the constraints on
        # matching pool file contents are retained.
        self.assert_collection_has_constraints(
            self.collection,
            [
                {
                    "collection_item_id": collection_item.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/hello_1.0.dsc",
                    "value": f"sha256:{dsc_file.hash_digest.hex()}",
                },
                {
                    "collection_item_id": collection_item.id,
                    "constraint_name": "pool-file",
                    "key": "pool/main/h/hello/hello_1.0.tar.xz",
                    "value": f"sha256:{tar_file.hash_digest.hex()}",
                },
            ],
        )

    def test_do_remove_item_artifact_reuse_versions(self) -> None:
        """If `may_reuse_versions=True`, versions may be reused."""
        self.collection.data["may_reuse_versions"] = True
        source_package_artifact = self.create_source_package(
            "hello", "1.0", ["hello_1.0.dsc", "hello_1.0.tar.xz"]
        )

        collection_item = self.manager.add_artifact(
            source_package_artifact,
            user=self.user,
            variables={"component": "main", "section": "devel"},
        )
        self.manager.remove_item(collection_item, user=self.user)

        collection_item.refresh_from_db()

        # The artifact is not removed yet (retention period applies)
        self.assertEqual(collection_item.artifact, source_package_artifact)

        self.assertEqual(collection_item.removed_by_user, self.user)
        self.assertIsInstance(collection_item.removed_at, dt.datetime)

        # Constraints on matching pool file contents are gone.
        self.assert_collection_has_constraints(self.collection, [])

    def test_lookup_unexpected_format_raise_lookup_error(self) -> None:
        """`lookup` raises `LookupError`: invalid format."""
        with self.assertRaisesRegex(
            LookupError, '^Unexpected lookup format: "binary:hello"$'
        ):
            self.manager.lookup("binary:hello")

    def test_lookup_return_none(self) -> None:
        """`lookup` returns None if there are no matches."""
        self.assertIsNone(self.manager.lookup("source:hello"))
        self.assertIsNone(self.manager.lookup("binary:hello_amd64"))
        self.assertIsNone(self.manager.lookup("index:missing"))
        self.assertIsNone(self.manager.lookup("name:nonexistent"))

    def test_lookup_source(self) -> None:
        """`lookup` returns a matching source package item."""
        items: list[CollectionItem] = []
        for name, version in (
            ("hello", "1.0"),
            ("hello", "1.1"),
            ("base-files", "1.0"),
        ):
            source_package_artifact = self.create_source_package(
                name, version, []
            )
            items.append(
                self.manager.add_artifact(
                    source_package_artifact,
                    user=self.user,
                    variables={"component": "main", "section": "devel"},
                )
            )

        source_package_artifact = self.create_source_package("hello", "1.2", [])
        items.append(
            self.manager.add_artifact(
                source_package_artifact,
                user=self.user,
                variables={"component": "main", "section": "devel"},
            )
        )
        self.manager.remove_item(items[-1], user=self.user)

        # CollectionItem of type BARE should not exist in this collection
        # (the manager does not allow adding it).  Add one to ensure that it
        # is filtered out.
        CollectionItem.objects.create(
            child_type=CollectionItem.Types.BARE,
            created_by_user=self.user,
            parent_collection=self.collection,
            category=ArtifactCategory.SOURCE_PACKAGE,
            name="something",
            data={"package": "hello", "version": "1.0"},
        )

        # items[0] (hello_1.0) and items[1] (hello_1.1) both match, but
        # items[1] has a higher version.  items[3] (hello_1.2) has been
        # removed.
        self.assertEqual(self.manager.lookup("source:hello"), items[1])

        self.assertEqual(
            self.manager.lookup("source-version:hello_1.0"), items[0]
        )
        self.assertEqual(
            self.manager.lookup("source-version:hello_1.1"), items[1]
        )
        self.assertIsNone(self.manager.lookup("source-version:hello_1.2"))

        self.assertEqual(self.manager.lookup("name:hello_1.0"), items[0])
        self.assertEqual(self.manager.lookup("name:hello_1.1"), items[1])
        self.assertEqual(self.manager.lookup("name:base-files_1.0"), items[2])
        self.assertIsNone(self.manager.lookup("name:hello_1.2"))

    def test_lookup_binary(self) -> None:
        """`lookup` returns a matching binary package item."""
        items: list[CollectionItem] = []
        for srcpkg_name, srcpkg_version, name, version, architecture in (
            ("hello", "1.0", "libhello1", "1.0", "amd64"),
            ("hello", "1.0", "libhello1", "1.0", "s390x"),
            ("hello", "1.0", "libhello-doc", "1.0", "all"),
            ("hello", "1.1", "libhello1", "1.1", "amd64"),
            ("base-files", "1.0", "base-files", "1:1.0", "amd64"),
        ):
            binary_package_artifact = self.create_binary_package(
                srcpkg_name, srcpkg_version, name, version, architecture, []
            )
            items.append(
                self.manager.add_artifact(
                    binary_package_artifact,
                    user=self.user,
                    variables={
                        "component": "main",
                        "section": "devel",
                        "priority": "optional",
                    },
                )
            )

        binary_package_artifact = self.create_binary_package(
            "hello", "1.2", "libhello1", "1.2", "amd64", []
        )
        items.append(
            self.manager.add_artifact(
                binary_package_artifact,
                user=self.user,
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
            )
        )
        self.manager.remove_item(items[-1], user=self.user)

        # CollectionItem of type BARE should not exist in this collection
        # (the manager does not allow adding it).  Add one to ensure that it
        # is filtered out.
        CollectionItem.objects.create(
            child_type=CollectionItem.Types.BARE,
            created_by_user=self.user,
            parent_collection=self.collection,
            category=ArtifactCategory.BINARY_PACKAGE,
            name="something",
            data={
                "package": "libhello1",
                "version": "1.0",
                "architecture": "amd64",
            },
        )

        # items[0] (libhello1_1.0_amd64) and items[3] (libhello_1.1_amd64)
        # both match, but items[3] has a higher version.  items[5]
        # (libhello1_1.2_amd64) has been removed.
        self.assertEqual(
            self.manager.lookup("binary:libhello1_amd64"), items[3]
        )

        self.assertEqual(
            self.manager.lookup("binary-version:libhello1_1.0_amd64"), items[0]
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello1_1.0_s390x"), items[1]
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello-doc_1.0_all"), items[2]
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello-doc_1.0_amd64"),
            items[2],
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello-doc_1.0_s390x"),
            items[2],
        )
        self.assertEqual(
            self.manager.lookup("binary-version:libhello1_1.1_amd64"), items[3]
        )
        self.assertIsNone(
            self.manager.lookup("binary-version:libhello1_1.1_s390x")
        )
        self.assertIsNone(
            self.manager.lookup("binary-version:libhello1_1.2_amd64")
        )

        self.assertEqual(
            self.manager.lookup("name:libhello1_1.0_amd64"), items[0]
        )
        self.assertEqual(
            self.manager.lookup("name:libhello1_1.0_s390x"), items[1]
        )
        self.assertEqual(
            self.manager.lookup("name:libhello-doc_1.0_all"), items[2]
        )
        self.assertEqual(
            self.manager.lookup("name:libhello1_1.1_amd64"), items[3]
        )
        self.assertEqual(
            self.manager.lookup("name:base-files_1:1.0_amd64"), items[4]
        )
        self.assertIsNone(self.manager.lookup("name:libhello1_1.2_amd64"))

    def test_lookup_repository_index(self) -> None:
        """``lookup`` returns a matching repository index."""
        items: list[CollectionItem] = []
        for path in (
            "main/source/Sources.xz",
            "main/binary-amd64/Packages.xz",
            "Release",
        ):
            index_artifact = self.playground.create_repository_index(
                PurePath(path).name, skip_add_files_in_store=True
            )
            items.append(
                self.manager.add_artifact(
                    index_artifact, user=self.user, variables={"path": path}
                )
            )

        index_artifact = self.playground.create_repository_index(
            "Packages.xz", skip_add_files_in_store=True
        )
        items.append(
            self.manager.add_artifact(
                index_artifact,
                user=self.user,
                variables={"path": "main/binary-amd64/Packages.xz"},
                replace=True,
            )
        )

        # CollectionItem of type BARE should not exist in this collection
        # (the manager does not allow adding it).  Add one to ensure that it
        # is filtered out.
        CollectionItem.objects.create(
            child_type=CollectionItem.Types.BARE,
            created_by_user=self.user,
            parent_collection=self.collection,
            category=ArtifactCategory.REPOSITORY_INDEX,
            name="something",
            data={"path": "Release"},
        )

        self.assertEqual(
            self.manager.lookup("index:main/source/Sources.xz"), items[0]
        )
        # items[1] and items[3] would both match, but only items[3] is
        # active.
        self.assertEqual(
            self.manager.lookup("index:main/binary-amd64/Packages.xz"), items[3]
        )
        self.assertEqual(self.manager.lookup("index:Release"), items[2])
        self.assertIsNone(self.manager.lookup("index:missing"))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-client guided setup interface."""

import json
import os
import platform
import re
import secrets
import urllib.parse
from pathlib import Path
from typing import Any, assert_never, override

import requests
import rich.box
from configobj import ConfigObj
from rich.console import Console
from rich.prompt import Confirm
from rich.table import Table
from xkcdpass import xkcd_password

from debusine.client.config import ConfigHandler, KNOWN_SERVERS, ServerInfo
from debusine.client.dataentry import DataEntry
from debusine.client.models import (
    EnrollConfirmPayload,
    EnrollOutcome,
    EnrollPayload,
)
from debusine.utils import get_user_agent
from debusine.utils.urls import URLPath


class TokenFetchFailed(Exception):
    """Exception raised when fetching a token failed."""


def fetch_token(console: Console, server: ServerInfo) -> str | None:
    """Get an API token from the server."""
    # Load data for generating challenges
    xkcd_wordfile = xkcd_password.locate_wordfile()
    xkcd_words = xkcd_password.generate_wordlist(
        wordfile=xkcd_wordfile, min_length=5, max_length=8
    )

    payload = EnrollPayload(
        # Nonce used to identify the enrolling session
        nonce=secrets.token_urlsafe(16),
        # Challenge text
        challenge=xkcd_password.generate_xkcdpassword(xkcd_words),
        scope=server.scope,
        hostname=platform.node(),
    )

    # URL used to get the token
    api_url_parsed = urllib.parse.urlparse(server.api_url)
    api_url_path = URLPath(api_url_parsed.path, is_file=False)
    api_url_path += "enroll/"
    enroll_url_parsed = api_url_parsed._replace(path=str(api_url_path))
    enroll_url = urllib.parse.urlunparse(enroll_url_parsed)

    # URL used for the user to confirm enrollment
    confirm_url_path = URLPath(api_url_parsed.path, is_file=False)
    confirm_url_path.pop()
    confirm_url_path += f"/-/enroll/confirm/{payload.nonce}"
    confirm_url_parsed = api_url_parsed._replace(path=str(confirm_url_path))
    confirm_url = urllib.parse.urlunparse(confirm_url_parsed)

    console.print(
        ":point_right:",
        "Please visit",
        f"[link={confirm_url}]{rich.markup.escape(confirm_url)}[/]",
        "to confirm registration",
    )
    console.print(
        ":point_right:",
        "Make sure the page mentions the passphrase:"
        f' "[bold]{payload.challenge}[/]"',
    )

    payload_dict = payload.model_dump()
    request_timeout = server.http_timeout

    try:
        headers = {"User-Agent": get_user_agent()}
        response = requests.post(
            enroll_url,
            json=payload_dict,
            headers=headers,
            timeout=request_timeout,
        )
    except Exception as e:
        raise TokenFetchFailed(f"request to server failed: {e}") from e

    if response.status_code != 200:
        lead = f"The server returned status code {response.status_code}: "
        try:
            result = response.json()
        except requests.exceptions.JSONDecodeError:
            raise TokenFetchFailed(lead + response.text)
        else:
            if message := result.get("title"):
                raise TokenFetchFailed(lead + message)
            else:
                raise TokenFetchFailed(lead + json.dumps(result))

    try:
        result = response.json()
    except requests.exceptions.JSONDecodeError:
        raise TokenFetchFailed("Invalid JSON in response")

    try:
        confirmation = EnrollConfirmPayload.model_validate(result)
    except ValueError:
        raise TokenFetchFailed("Invalid response payload")

    match confirmation.outcome:
        case EnrollOutcome.CONFIRM:
            console.print(
                ":white_check_mark: Confirmation confirmed, token acquired"
            )
            return confirmation.token
        case EnrollOutcome.CANCEL:
            console.print(":exclamation_mark: Confirmation cancelled")
            return None
        case _ as unreachable:
            assert_never(unreachable)


class DebusineClientConfig:
    """Editable configuration for debusine client."""

    def __init__(self, path: Path) -> None:
        """
        Initialize from a path.

        :param path: path to the configuration file
        """
        self.path = path
        self.config = ConfigObj(
            self.path.as_posix(),
            interpolation=False,
            list_values=False,
        )
        self.servers: dict[str, ServerInfo] = {
            name[7:]: ServerInfo.from_config(name[7:], section, path)
            for name, section in self.config.items()
            if name.startswith("server:")
        }
        self.edited: dict[str, ServerInfo] = {}

        # Currently selected server
        self.current_server: ServerInfo | None = None

        # Name of the server set as default
        self.default_server: str | None = None
        if section := self.config.get("General"):
            self.default_server = section.get("default-server")

    def select(self, name: str, desc: str) -> None:
        """
        Select the server to edit.

        :param name: server name
        :param desc: optional description used if creating a new server
        """
        if server := self.edited.get(name):
            # Server is already being edited
            pass
        elif server := self.servers.get(name):
            # Server is in the existing configuration
            self.edited[server.name] = server
        elif server := KNOWN_SERVERS.get(name):
            # Server is known
            self.edited[server.name] = server
        else:
            server = ServerInfo.from_string(name, desc)
            self.edited[server.name] = server
        self.current_server = server

    @property
    def server(self) -> ServerInfo:
        """Require the current server to be set and return it."""
        if self.current_server is None:
            raise RuntimeError("current server has not been selected")
        return self.current_server

    def set_api_url(self, api_url: str) -> None:
        """Set the API url of the current server."""
        edited = self.server._replace(api_url=api_url)
        self.current_server = edited
        self.edited[edited.name] = edited

    def set_scope(self, scope: str) -> None:
        """Set the scope of the current server."""
        edited = self.server._replace(scope=scope)
        self.current_server = edited
        self.edited[edited.name] = edited

    def set_api_token(self, api_token: str | None) -> None:
        """Set the API token of the current server."""
        edited = self.server._replace(api_token=api_token)
        self.current_server = edited
        self.edited[edited.name] = edited

    def set_default(self) -> None:
        """Set the current server as default server."""
        self.default_server = self.server.name

    def set_default_workspace(self, workspace: str) -> None:
        """Set the default workspace of the current server."""
        edited = self.server._replace(default_workspace=workspace)
        self.current_server = edited
        self.edited[edited.name] = edited

    def fetch_token(self, console: Console) -> bool:
        """Fetch a token and set it in the current configuration."""
        try:
            api_token = fetch_token(console, self.server)
        except TokenFetchFailed as e:
            console.print(":exclamation_mark:", rich.markup.escape(str(e)))
            return False
        self.set_api_token(api_token)
        return True

    def save(self) -> None:
        """Save the configuration."""
        for name, current in self.edited.items():
            orig = self.servers.get(name)

            # Get the server section to save
            key = f"server:{name}"
            if (section := self.config.get(key)) is None:
                self.config[key] = {}
                section = self.config[key]

            # Update with the edited configuration
            if orig is None or orig.api_url != current.api_url:
                section["api-url"] = current.api_url
            if orig is None or orig.scope != current.scope:
                section["scope"] = current.scope
            if orig is None or orig.api_token != current.api_token:
                if current.api_token is None:
                    if "token" in section:
                        del section["token"]
                else:
                    section["token"] = current.api_token
            if (
                orig is None
                or orig.default_workspace != current.default_workspace
            ):
                section["default-workspace"] = current.default_workspace

        if self.default_server is not None:
            if section := self.config.get("General"):
                if section.get("default-server") != self.default_server:
                    section["default-server"] = self.default_server
            else:
                self.config["General"] = {"default-server": self.default_server}

        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.config.write()


class ServerConfigEditor(DataEntry):
    """Interactive editor for ServerConfig."""

    prompt = "> "

    config: DebusineClientConfig

    def __init__(
        self,
        config: DebusineClientConfig,
        console: Console | None = None,
    ) -> None:
        """Initialize the ServerConfig editor."""
        super().__init__(console=console)
        self.config = config

    def lint_message(self, message: str) -> None:
        """Print a linter message."""
        self.console.print(f":point_right: {message}", highlight=False)

    @override
    def menu(self) -> None:
        server = self.config.server
        is_default = self.config.server.name == self.config.default_server
        title = f"Configuration for {server.name}"
        if is_default:
            title += " [green not bold](default server)[/]"
        else:
            title += (
                " [yellow not bold]"
                f"(default is [i]{self.config.default_server}[/])"
                "[/]"
            )
        table = Table(box=rich.box.SIMPLE)
        table.add_column(title, justify="left", no_wrap=True)
        table.add_column("", justify="left", no_wrap=True)

        table.add_row(rf"\[server:[bold]{server.name}[bold]]", "")
        table.add_row(f"scope = [bold]{server.scope}[bold]", "([bold]sc[/]ope)")
        table.add_row(
            f"api-url = [bold]{server.api_url}[bold]", "([bold]u[/]rl)"
        )
        table.add_row(
            "token = [italic]"
            + ("acquired on save" if server.api_token is None else "present")
            + "[/]",
            "([bold]t[/]oken)",
        )
        table.add_row(
            f"default-workspace = [bold]{server.default_workspace}[bold]",
            "([bold]w[/]orkspace)",
        )

        self.console.print(table)

        for msg in server.lint():
            self.lint_message(msg)
        self.console.print()
        self.console.print(
            "Commands: [bold]u[/]rl, [bold]t[/]oken, [bold]sc[/]ope, "
            "(make) [bold]d[/]efault, [bold]s[/]ave, [bold]q[/]uit:"
        )

    def do_default(self, arg: str) -> None:  # noqa: ARG002, U100
        """Set as default server."""
        self.config.set_default()

    def do_scope(self, arg: str) -> None:  # noqa: ARG002, U100
        """Change the scope."""
        server = self.config.server
        self.config.set_scope(
            self.input_line("Enter scope", initial=server.scope, required=True)
        )

    def do_url(self, arg: str) -> None:  # noqa: ARG002, U100
        """Change the API URL."""
        server = self.config.server
        self.config.set_api_url(
            self.input_line(
                "Enter API URL", initial=server.api_url, required=True
            )
        )

    def do_token(self, arg: str) -> None:  # noqa: ARG002, U100
        """Change the token."""
        server = self.config.server
        self.config.set_api_token(
            self.input_line(
                "Enter (or clear) API token", initial=server.api_token
            )
            or None
        )

    def do_workspace(self, arg: str) -> None:  # noqa: ARG002, U100
        """Change the default workspace."""
        server = self.config.server
        self.config.set_default_workspace(
            self.input_line(
                "Enter default workspace name",
                initial=server.default_workspace,
                required=True,
            )
        )

    @override
    def do_quit(self, arg: str) -> bool:
        self.console.print(
            ":hand_with_fingers_splayed: Configuration left unchanged"
        )
        return super().do_quit(arg)

    def do_save(self, arg: str) -> bool:  # noqa: ARG002, U100
        """Save the new configuration."""
        if not self.config.server.api_token:
            while True:
                if self.config.fetch_token(self.console):
                    break
                retry = Confirm.ask(
                    "Fetch failed: retry?", console=self.console
                )
                if not retry:
                    break

        self.config.save()
        self.console.print(":thumbs_up: Configuration saved")
        return True


class ServerSelector(DataEntry):
    """Interactive server selector."""

    prompt = "> "

    entries: list[ServerInfo]
    selected: ServerInfo | None

    def __init__(
        self,
        entries: list[ServerInfo],
        console: Console | None = None,
        default: str | None = None,
    ) -> None:
        """Initialize the Server selector."""
        super().__init__(console=console)
        self.entries = entries
        self.default_name = default
        self.selected = None

    @override
    def menu(self) -> None:
        """Show a list of entries to pick from."""
        table = Table(title="Configured servers", box=rich.box.SIMPLE)
        table.add_column("#", justify="right", no_wrap=True)
        table.add_column("Name", justify="left", no_wrap=True)
        table.add_column("Current", justify="left", no_wrap=True)
        table.add_column("Description", justify="left", no_wrap=True)

        for idx, info in enumerate(self.entries, start=1):
            table.add_row(
                str(idx),
                info.name,
                "yes" if info.name == self.default_name else "",
                info.desc,
            )

        self.console.print(table)

        self.console.print()

        if self.entries:
            if len(self.entries) == 1:
                numbers = "[bold]1[/], "
            else:
                numbers = f"[bold]1…{len(self.entries)}[/], "
        else:
            numbers = ""

        self.console.print(
            f"{numbers}[bold]n[/]ew, [bold]q[/]uit:", highlight=False
        )

    def do_new(self, arg: str) -> bool:  # noqa: ARG002, U100
        """Create a new entry."""
        entry = self.input_line("New server name or URL")
        if not entry:
            return False
        else:
            self.selected = ServerInfo.from_string(
                entry, desc="Manually entered"
            )
            return True

    # Cmd.default is really expected to return a bool|None, and it looks like
    # badly typed (see how Cmd.onecmd uses it)
    @override
    def default(self, line: str) -> bool:  # type: ignore[override]
        """Handle entry IDs."""
        line = line.strip()
        if not re.match("^[0-9]+$", line):
            return super().default(line)

        value = int(line)
        if value < 1 or value > len(self.entries):
            self.console.print(
                "[red]Entry number must be between 1 and"
                f" {len(self.entries)}[/]"
            )
            return False

        self.selected = self.entries[value - 1]
        return True

    @override
    def cmdloop(self, intro: Any | None = None) -> None:
        """Run data entry."""
        if not self.entries:
            self.do_new("")
        else:
            super().cmdloop(intro=intro)


def setup_server(
    config_file_path: (
        str | os.PathLike[str]
    ) = ConfigHandler.DEFAULT_CONFIG_FILE_PATH,
    server: str | None = None,
    scope: str | None = None,
    console: Console | None = None,
) -> None:
    """Create or edit a server entry."""
    console = console or Console()
    config = DebusineClientConfig(Path(config_file_path))

    if server is None:
        # Ask the user to pick a server
        servers: list[ServerInfo] = []
        servers.extend(config.servers.values())
        for info in KNOWN_SERVERS.values():
            if info.name in config.servers:
                continue
            servers.append(info)
        server_selector = ServerSelector(
            servers, console=console, default=config.default_server
        )
        server_selector.cmdloop()
        if server_selector.selected is None:
            # Abort if no server was selected
            return
        config.select(server_selector.selected.name, "selected from menu")
    else:
        config.select(server, "provided on command line")

    if scope is not None:
        config.set_scope(scope)

    # Set server as default if no default was set
    if config.default_server is None:
        config.set_default()
        console.print(
            ":star: Default server was not set,"
            f" setting it to [bold]{config.default_server}[/]"
        )

    server_config_editor = ServerConfigEditor(config, console=console)
    server_config_editor.cmdloop()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to delete a notification channel."""

from typing import Any, NoReturn, override

from django.core.management import CommandError, CommandParser

from debusine.db.models import NotificationChannel
from debusine.django.management.debusine_base_command import DebusineBaseCommand


class Command(DebusineBaseCommand):
    """Command to delete a notification channel."""

    help = "Delete a notification channel"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument("name", help="Notification channel name")

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Delete NotificationChannel."""
        name = options["name"]

        try:
            notification_channel = NotificationChannel.objects.get(name=name)
        except NotificationChannel.DoesNotExist:
            raise CommandError(
                f'NotificationChannel "{name}" does not exist', returncode=3
            )

        notification_channel.delete()

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used for file backend configuration."""

import pydantic


class FileBackendConfiguration(pydantic.BaseModel):
    """Base class for file backend configuration."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")


class LocalFileBackendConfiguration(FileBackendConfiguration):
    """Configuration for `LocalFileBackend`."""

    base_directory: str | None = None


class ExternalDebianSuiteBackendConfiguration(FileBackendConfiguration):
    """Configuration for `ExternalDebianSuiteFileBackend`."""

    archive_root_url: str
    suite: str
    components: list[str]


class MemoryFileBackendConfiguration(FileBackendConfiguration):
    """Configuration for `MemoryFileBackend`."""

    name: str


class S3FileBackendConfiguration(FileBackendConfiguration):
    """Configuration for `S3FileBackend`."""

    bucket_name: str
    # https://docs.aws.amazon.com/AmazonS3/latest/userguide/storage-class-intro.html
    storage_class: str | None = None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the UnshareExecutor class."""

import os
import shlex
import stat
from pathlib import Path, PurePath
from shutil import rmtree
from subprocess import PIPE
from typing import Any, override
from unittest import mock
from unittest.mock import MagicMock, call

import psutil

from debusine.client.debusine import Debusine
from debusine.tasks import Noop
from debusine.tasks.executors.base import (
    ExecutorStatistics,
    ImageNotDownloadedError,
    InstanceNotRunning,
    executor_class,
)
from debusine.tasks.executors.unshare import UnshareExecutor, UnshareInstance
from debusine.tasks.models import BackendType
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase


class UnshareExecutorTests(ExternalTaskHelperMixin[Noop], TestCase):
    """Unit tests for UnshareExecutor."""

    @override
    def setUp(self) -> None:
        """Mock the Debusine API for tests."""
        super().setUp()
        self.debusine_api = MagicMock(spec=Debusine)
        self.image_artifact = self.mock_image_download(self.debusine_api)

        self.workdir = self.create_temporary_directory()
        self.executor = UnshareExecutor(
            self.debusine_api,
            42,
            workspace="workspace",
            working_directory=self.workdir,
        )

    def test_backend_name(self) -> None:
        """Test that the backend_name attribute was set."""
        self.assertEqual(UnshareExecutor.backend_name, BackendType.UNSHARE)

    def test_available(self) -> None:
        """Test that available() returns True if unshare is available."""
        self.mock_is_command_available({"unshare": True, "ip": True})
        self.assertTrue(UnshareExecutor.available())

    def test_available_no_unshare(self) -> None:
        """Test that available() returns False if unshare is not available."""
        self.mock_is_command_available({"unshare": False, "ip": True})
        self.assertFalse(UnshareExecutor.available())

    def test_available_no_iproute2(self) -> None:
        """Test that available() returns False if unshare is not available."""
        self.mock_is_command_available({"unshare": True, "ip": False})
        self.assertFalse(UnshareExecutor.available())

    def test_instantiation_fetches_artifact(self) -> None:
        """Test that instantiating UnshareExecutor fetches the artifact."""
        self.assertEqual(self.executor.system_image, self.image_artifact)

    def test_download_image(self) -> None:
        """Test that download_image calls ImageCache.download_image."""
        response = self.executor.download_image()
        expected_path = self.image_cache_path / "42/system.tar.xz"
        self.assertEqual(self.executor._local_path, expected_path)
        self.assertEqual(response, str(expected_path))

    def test_image_name(self) -> None:
        """Test that image_name returns the previously downloaded image path."""
        self.executor.download_image()
        response = self.executor.image_name()
        expected_path = self.image_cache_path / "42/system.tar.xz"
        self.assertEqual(response, str(expected_path))

    def test_image_name_not_downloaded(self) -> None:
        """Test that image_name fails if there is no name."""
        self.assertRaises(ImageNotDownloadedError, self.executor.image_name)

    def test_autopkgtest_virt_server(self) -> None:
        """Test that autopkgtest_virt_server returns unshare."""
        self.assertEqual(self.executor.autopkgtest_virt_server(), "unshare")

    def test_autopkgtest_virt_args(self) -> None:
        """Test that autopkgtest_virt_args returns sane arguments."""
        self.executor._local_path = Path("/not/used/path/system.tar.xz")
        self.assertEqual(
            self.executor.autopkgtest_virt_args(),
            [
                "--arch",
                "amd64",
                "--release",
                "bookworm",
                "--tarball",
                str(self.executor._local_path),
            ],
        )

    def test_autopkgtest_virt_args_before_download(self) -> None:
        """Test that autopkgtest_virt_args requires download."""
        self.assertRaises(
            ImageNotDownloadedError, self.executor.autopkgtest_virt_args
        )

    def test_executor_class_finds_unshare(self) -> None:
        """Test that executor_class() supports unshare."""
        instance = executor_class(BackendType.UNSHARE)
        self.assertEqual(instance, UnshareExecutor)

    def test_create(self) -> None:
        """Test create() return UnshareInstance instance."""
        self.executor._local_path = self.create_temporary_file()
        unshare_instance = self.executor.create()

        self.assertIsInstance(unshare_instance, UnshareInstance)

        self.assertEqual(unshare_instance._image, self.executor._local_path)

    def test_create_raise_runtime_error(self) -> None:
        """Test create() raise ImageNotDownloadedError: _local_path is None."""
        msg_starts_with = r"^UnshareExecutor\.download_image\(\) should have"
        with self.assertRaisesRegex(ImageNotDownloadedError, msg_starts_with):
            self.executor.create()

    def test_get_statistics_without_instance(self) -> None:
        """`get_statistics` returns some statistics without an instance."""
        memory = psutil.virtual_memory()

        with mock.patch("psutil.virtual_memory", return_value=memory):
            statistics = UnshareExecutor.get_statistics(None)

        self.assertEqual(
            statistics,
            ExecutorStatistics(
                memory=memory.used,
                available_memory=memory.total,
                cpu_count=psutil.cpu_count() or 1,
            ),
        )

    def test_get_statistics_with_instance(self) -> None:
        """`get_statistics` returns all statistics with an instance."""
        image_file = self.create_temporary_file()
        working_directory = self.create_temporary_directory()
        instance = UnshareInstance(image_file, working_directory)
        instance._extracted = working_directory
        instance._root.mkdir()
        statvfs = os.statvfs(instance._root)
        memory = psutil.virtual_memory()

        with (
            mock.patch("psutil.virtual_memory", return_value=memory),
            mock.patch("os.statvfs", return_value=statvfs),
        ):
            statistics = UnshareExecutor.get_statistics(instance)

        self.assertEqual(
            statistics,
            ExecutorStatistics(
                disk_space=(
                    (statvfs.f_blocks - statvfs.f_bfree) * statvfs.f_frsize
                ),
                memory=memory.used,
                available_disk_space=statvfs.f_blocks * statvfs.f_frsize,
                available_memory=memory.total,
                cpu_count=psutil.cpu_count() or 1,
            ),
        )

    def test_get_statistics_with_terminated_instance(self) -> None:
        """`get_statistics` handles an instance that has already exited."""
        image_file = self.create_temporary_file()
        working_directory = self.create_temporary_directory()
        instance = UnshareInstance(image_file, working_directory)
        instance._extracted = working_directory
        memory = psutil.virtual_memory()

        with (
            mock.patch("psutil.virtual_memory", return_value=memory),
        ):
            statistics = UnshareExecutor.get_statistics(instance)

        self.assertEqual(
            statistics,
            ExecutorStatistics(
                memory=memory.used,
                available_memory=memory.total,
                cpu_count=psutil.cpu_count() or 1,
            ),
        )


class UnshareInstanceTests(TestCase):
    """Tests for UnshareInstance class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.image_file = self.create_temporary_file()
        self.working_directory = self.create_temporary_directory()

        self.instance = UnshareInstance(self.image_file, self.working_directory)

        (
            self.run_unshare_command_patcher,
            self.run_unshare_command_mocked,
        ) = self.patch_run_unshare_command()

    @override
    def tearDown(self) -> None:
        """Remove any unpacked instances during tests."""
        if self.instance._extracted:
            rmtree(self.instance._extracted)
        super().tearDown()

    def patch_run_unshare_command(self) -> tuple[Any, MagicMock]:
        """Patch _run_share_cmd() and return its patcher and mock."""
        patcher = mock.patch(
            "debusine.tasks.executors.unshare."
            "UnshareInstance._run_unshare_command"
        )
        mocked = patcher.start()

        self.addCleanup(patcher.stop)
        return patcher, mocked

    def test_start(self) -> None:
        """start() uncompress the image."""
        _, run_unshare_command_mocked = self.patch_run_unshare_command()

        self.instance._image = MagicMock(spec=Path)

        self.instance.start()

        assert self.instance._extracted is not None
        self.assertEqual(
            stat.S_IMODE(self.instance._extracted.stat().st_mode), 0o755
        )

        root = str(self.instance._root)

        image_file_open_context = self.instance._image.open(
            "rb"
        ).__enter__.return_value

        expected_calls = [
            call(["--map-root-user", "chown", "1:1", root]),
            call(
                [
                    "--setuid=0",
                    "--setgid=0",
                    "tar",
                    "-C",
                    root,
                    "--xz",
                    "-x",
                ],
                stdin=image_file_open_context,
            ),
        ]

        run_unshare_command_mocked.assert_has_calls(expected_calls)

    def test_stop(self) -> None:
        """stop() delete the contents using unshare."""
        self.instance.start()

        extracted = self.instance._extracted
        assert extracted is not None
        root = self.instance._root

        self.instance.stop()

        self.run_unshare_command_mocked.assert_called_with(
            ["--map-root-user", "rm", "-rf", str(root)]
        )

        self.assertIsNone(self.instance._extracted)
        self.assertFalse(extracted.exists())

    def test_restart(self) -> None:
        """restart() is noop for a running instance."""
        self.instance.start()
        self.instance.restart()

    def test_do_file_push(self) -> None:
        """do_file_push() call _run_unshare_command."""
        self.instance.start()

        uid = 0
        gid = 101

        target = PurePath("/tmp/some_file.txt")

        inner_file = self.instance._root / target.relative_to("/")

        inner_parent_directory = inner_file.parent
        inner_parent_directory.mkdir(parents=True)

        source = MagicMock(spec=Path)
        source_open_context = source.open("rb").__enter__.return_value

        self.instance.do_file_push(source, target, uid, gid, mode=0o644)

        expected_args = [
            f"--setuid={uid}",
            f"--setgid={gid}",
            "sh",
            "-c",
            f"cat > {inner_file} && chmod 644 {inner_file}",
        ]

        self.run_unshare_command_mocked.assert_called_with(
            expected_args, stdin=source_open_context, check=True
        )

    def test_file_pull(self) -> None:
        """file_pull() copies the file."""
        self.instance.start()

        source = PurePath("/etc/hosts")
        target = self.create_temporary_file()

        self.instance.file_pull(source, target)

        self.run_unshare_command_mocked.assert_called_with(
            [
                "--setuid=0",
                "--setgid=0",
                "--map-root-user",
                "cp",
                "--no-preserve=ownership,mode",
                str(self.instance._root / source.relative_to("/")),
                str(target),
            ],
            check=True,
        )

    def test_run_as_non_root(self) -> None:
        """run(): call _run_unshare_command and return its return value."""
        self.instance.start()
        cmd = ["lintian", "--info"]
        self.instance._uid_map["_debusine"] = 42
        actual = self.instance.run(cmd)
        root = shlex.quote(str(self.instance._root))
        self.run_unshare_command_mocked.assert_has_calls(
            [
                call(
                    [
                        "--setuid=0",
                        "--setgid=0",
                        "--kill-child",
                        "--pid",
                        "--mount",
                        f"--mount-proc={self.instance._root}/proc",
                        "sh",
                        "-c",
                        f"mount --rbind /dev {root}/dev && "
                        f"exec /usr/sbin/chroot {root} "
                        f"/sbin/runuser -u _debusine -- {' '.join(cmd)}",
                    ],
                    stderr=PIPE,
                    stdout=PIPE,
                    text=None,
                ),
            ],
        )

        self.assertEqual(actual, self.run_unshare_command_mocked.return_value)

    def test_run_as_root(self) -> None:
        """run(): call _run_unshare_command as root and return."""
        self.instance.start()
        cmd = ["lintian", "--info"]
        actual = self.instance.run(cmd, run_as_root=True)
        root = shlex.quote(str(self.instance._root))
        self.run_unshare_command_mocked.assert_has_calls(
            [
                call(
                    [
                        "--setuid=0",
                        "--setgid=0",
                        "--kill-child",
                        "--pid",
                        "--mount",
                        f"--mount-proc={self.instance._root}/proc",
                        "sh",
                        "-c",
                        f"mount --rbind /dev {root}/dev && "
                        f"exec /usr/sbin/chroot {root} "
                        f"{' '.join(cmd)}",
                    ],
                    stderr=PIPE,
                    stdout=PIPE,
                    text=None,
                ),
            ],
        )

        self.assertEqual(actual, self.run_unshare_command_mocked.return_value)

    def test_run_cwd(self) -> None:
        """Run a command in a different working directory."""
        self.instance.start()
        cmd = ["lintian", "--info"]
        self.instance._uid_map["_debusine"] = 42
        actual = self.instance.run(cmd, cwd="work&stuff")
        root = shlex.quote(str(self.instance._root))
        self.run_unshare_command_mocked.assert_has_calls(
            [
                call(
                    [
                        "--setuid=0",
                        "--setgid=0",
                        "--kill-child",
                        "--pid",
                        "--mount",
                        f"--mount-proc={self.instance._root}/proc",
                        "sh",
                        "-c",
                        f"mount --rbind /dev {root}/dev && "
                        f"exec /usr/sbin/chroot {root} "
                        f"/sbin/runuser -u _debusine -- "
                        # cwd is quoted as 'work&stuff' for the inner shell,
                        # but then this must be quoted again for the outer
                        # shell, resulting in each single quote being
                        # enclosed within separate double quotes.
                        f"sh -c 'cd '\"'\"'work&stuff'\"'\"' && "
                        f"{' '.join(cmd)}'",
                    ],
                    stderr=PIPE,
                    stdout=PIPE,
                    text=None,
                ),
            ],
        )

        self.assertEqual(actual, self.run_unshare_command_mocked.return_value)

    def patch_unshare_command_args(self, command: str, args: list[str]) -> None:
        """Patch UnshareExecutor._unshare_command and _unshare_default_args."""
        unshare_command = mock.patch(
            "debusine.tasks.executors.unshare.UnshareInstance.unshare_command",
            new=command,
        )
        unshare_command.start()
        self.addCleanup(unshare_command.stop)

        unshare_default_args = mock.patch(
            "debusine.tasks.executors.unshare."
            "UnshareInstance._unshare_default_args",
            new=args,
        )
        unshare_default_args.start()
        self.addCleanup(unshare_default_args.stop)

    def test_run_unshare_command(self) -> None:
        """run_unshare_command() runs a command, output is as expected."""
        self.run_unshare_command_patcher.stop()
        self.patch_unshare_command_args("echo", ["-n"])

        output = "test"
        completed = UnshareInstance._run_unshare_command([output], text=True)

        self.assertEqual(completed.stdout, output)
        self.assertEqual(completed.returncode, 0)

    def test_run_unshare_command_with_stderr_stdout(self) -> None:
        """run_unshare_command() runs a command with stderr and stdout."""
        self.run_unshare_command_patcher.stop()

        self.patch_unshare_command_args("sh", ["-c"])

        stderr_contents = "hello stderr"
        stdout_contents = "hello stdout"

        stderr_file_path = self.create_temporary_file()
        stdout_file_path = self.create_temporary_file()

        with (
            stderr_file_path.open("wb") as stderr_file,
            stdout_file_path.open("wb") as stdout_file,
        ):
            UnshareInstance._run_unshare_command(
                [
                    f"printf %s '{stderr_contents}' 1>&2; "
                    f"printf %s '{stdout_contents}'",
                ],
                stderr=stderr_file.fileno(),
                stdout=stdout_file.fileno(),
            )

        self.assertEqual(stderr_file_path.read_text(), stderr_contents)
        self.assertEqual(stdout_file_path.read_text(), stdout_contents)

    def test_run_unshare_command_environment(self) -> None:
        """run_unshare_command() runs a command in a sanitized environment."""
        self.run_unshare_command_patcher.stop()
        # Yes, this runs "env env", but "sh -c env" would (annoyingly) set PWD
        self.patch_unshare_command_args("env", [])

        completed = UnshareInstance._run_unshare_command(["env"], text=True)
        output = (
            "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
            "\n"
        )
        self.assertEqual(completed.stdout, output)
        self.assertEqual(completed.returncode, 0)

    def test_run_unshare_command_env_none(self) -> None:
        """env=None is the same as not specifying env."""
        self.run_unshare_command_patcher.stop()
        # Yes, this runs "env env", but "sh -c env" would (annoyingly) set PWD
        self.patch_unshare_command_args("env", [])

        completed = UnshareInstance._run_unshare_command(
            ["env"], text=True, env=None
        )
        output = (
            "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
            "\n"
        )
        self.assertIn(completed.stdout, output)
        self.assertEqual(completed.returncode, 0)

    def test_run_unshare_command_environment_override(self) -> None:
        """run_unshare_command() lets the caller specify an environment."""
        self.run_unshare_command_patcher.stop()
        self.patch_unshare_command_args("env", [])

        completed = UnshareInstance._run_unshare_command(
            ["env"], text=True, env={"FOO": "BAR"}
        )
        output = (
            "FOO=BAR\n"
            "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
            "\n"
        )
        self.assertEqual(completed.stdout, output)
        self.assertEqual(completed.returncode, 0)

    def test_root(self) -> None:
        """_root property returns self._extracted / root."""
        self.instance._extracted = self.create_temporary_directory()

        self.assertEqual(self.instance._root, self.instance._extracted / "root")

    def test_root_raise_runtime_error(self) -> None:
        """start() must be called before using _root property."""
        msg_starts_with = (
            r"^UnshareInstance.start\(\) must be called before using _root$"
        )
        with self.assertRaisesRegex(InstanceNotRunning, msg_starts_with):
            self.instance._root



# Generated by Django 4.2.20 on 2025-06-18 13:47

from django.db import migrations, models

import debusine.artifacts.models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0013_qa_suite'),
    ]
    replaces = [
        ('db', '0155_alter_workrequest_result'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workrequest',
            name='result',
            field=models.CharField(
                choices=[
                    (debusine.artifacts.models.WorkRequestResults['NONE'], ''),
                    (
                        debusine.artifacts.models.WorkRequestResults['SUCCESS'],
                        'Success',
                    ),
                    (
                        debusine.artifacts.models.WorkRequestResults['FAILURE'],
                        'Failure',
                    ),
                    (
                        debusine.artifacts.models.WorkRequestResults['ERROR'],
                        'Error',
                    ),
                    (
                        debusine.artifacts.models.WorkRequestResults['SKIPPED'],
                        'Skipped',
                    ),
                ],
                default=debusine.artifacts.models.WorkRequestResults['NONE'],
                editable=False,
                max_length=7,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Table ordering definitions."""

from collections.abc import Sequence
from functools import cached_property
from typing import Any, Optional, TYPE_CHECKING, assert_never, cast, override

from django.core.exceptions import ImproperlyConfigured
from django.db.models import F, Model, QuerySet
from django.db.models.expressions import OrderBy
from django.forms.utils import pretty_name
from django.utils.html import format_html
from django.utils.safestring import SafeString

from debusine.web.views.table.declarative import Attribute

if TYPE_CHECKING:
    from debusine.web.views.table.table import Table

type OrderBys = str | F | OrderBy | Sequence[str | F | OrderBy]


def to_order_by(val: str | F | OrderBy) -> OrderBy:
    """Convert the argument to an F-expression if needed."""
    match val:
        case str():
            if val.startswith("-"):
                return F(val[1:]).desc()
            else:
                return F(val).asc()
        case F():
            return val.asc()
        case OrderBy():
            return val
        case _ as unreachable:
            assert_never(unreachable)


def order_by_tuple(arg: OrderBys) -> tuple[OrderBy, ...]:
    """Normalize an argument to a tuple of OrderBy expressions."""
    match arg:
        case str() | F() | OrderBy():
            return (to_order_by(arg),)
        case _:
            return tuple(to_order_by(x) for x in arg)


class Sort:
    """A way to sort table rows."""

    #: Django ORM order_by expression
    order_by: tuple[OrderBy, ...]
    #: User-visible label
    label: str
    #: Icon name
    icon: str

    def __init__(
        self, order_by: OrderBys, label: str, icon: str = "sort-alpha-down"
    ) -> None:
        """
        Store sort information.

        :param order_by: ORM ordering definition
        :param label: label for this sort style
        :param icon: icon for this sort style
        """
        self.order_by = order_by_tuple(order_by)
        self.label = label
        self.icon = icon


class Ordering(Attribute):
    """
    Declarative definition of a bundle of ordering options.

    This is a shortcut for declaratively defining one or more related ordering
    variants.
    """

    #: order_by expressions indexed by ordering name
    options: dict[str, Sort]

    def __init__(
        self,
        asc: Sort | OrderBys | None = None,
        desc: Sort | OrderBys | None = None,
        label: str | None = None,
        **kwargs: Sort | OrderBys,
    ) -> None:
        """
        Parse arguments into Sort instances.

        :param asc: ascending or default sort order
        :param desc: descending/reverse of asc
        :param label: ordering label (optional, inferred if needed)
        """
        self.label = label
        self.options = {}
        if asc is not None:
            self.options["asc"] = (
                asc if isinstance(asc, Sort) else self.make_asc(asc)
            )
        if desc is not None:
            self.options["desc"] = (
                desc if isinstance(desc, Sort) else self.make_desc(desc)
            )
        elif asc is not None:
            self.options["desc"] = self.make_desc(
                tuple(
                    cast(OrderBy, x.copy().reverse_ordering())
                    for x in self.options["asc"].order_by
                )
            )
        for name, sort in kwargs.items():
            if not isinstance(sort, Sort):
                raise ImproperlyConfigured(
                    "ordering that isn't 'asc' or 'desc' needs"
                    " to be given as full 'Sort' instances"
                )
            self.options[name] = sort

    def make_asc(self, order_by: OrderBys) -> Sort:
        """Infer details for ascending sorting."""
        return Sort(
            order_by_tuple(order_by), "Alphabetical order", "sort-alpha-down"
        )

    def make_desc(self, order_by: OrderBys) -> Sort:
        """Infer details for descending sorting."""
        return Sort(
            order_by_tuple(order_by),
            "Reverse alphabetical order",
            "sort-alpha-up-alt",
        )


class StringOrdering(Ordering):
    """Sorts strings alphabetically."""


class NumberOrdering(Ordering):
    """Sort numbers."""

    @override
    def make_asc(self, order_by: OrderBys) -> Sort:
        return Sort(
            order_by_tuple(order_by), "Lowest first", "sort-numeric-down"
        )

    @override
    def make_desc(self, order_by: OrderBys) -> Sort:
        return Sort(
            order_by_tuple(order_by),
            "Highest first",
            "sort-numeric-up-alt",
        )


class DateTimeOrdering(Ordering):
    """Sort datetimes."""

    @override
    def make_asc(self, order_by: OrderBys) -> Sort:
        return Sort(
            order_by_tuple(order_by), "Oldest first", "sort-numeric-down"
        )

    @override
    def make_desc(self, order_by: OrderBys) -> Sort:
        return Sort(
            order_by_tuple(order_by),
            "Newest first",
            "sort-numeric-up-alt",
        )


def compute_sort_icon_current_next(
    orderings: list["TableOrdering[Any]"],
) -> tuple[Optional["TableOrdering[Any]"], Optional["TableOrdering[Any]"]]:
    """Select current and next orderings to use for sort icons."""
    if not orderings:
        return None, None
    elif len(orderings) == 1:
        return (
            orderings[0] if orderings[0].active else None,
            orderings[0] if not orderings[0].active else None,
        )
    else:
        # Cycle between available sort options
        for idx, ordering in enumerate(orderings):
            if ordering.active:
                return (
                    ordering,
                    orderings[(idx + 1) % len(orderings)],
                )
        else:
            return (None, orderings[0])


def render_sort_icon[M: Model](
    orderings: list["TableOrdering[M]"], for_table: bool = False
) -> SafeString:
    """Render an icon showing current ordering."""
    current, next_ = compute_sort_icon_current_next(orderings)
    class_ = "btn"
    if for_table:
        class_ += " btn-debusine-toolbar"
    else:
        class_ += " btn-sm"
    if current:
        if not for_table:
            class_ += " active"
        if next_:
            return format_html(
                "<a href='?{query_string}' class='{class_}'>"
                "<i class='bi bi-{icon}'"
                " title='{current_label} → {next_label}'>"
                "</i></a>",
                class_=SafeString(class_),
                query_string=next_.query_string,
                icon=current.icon,
                current_label=current.sort.label,
                next_label=next_.sort.label,
            )
        else:
            return format_html(
                "<span class='{class_}'>"
                "<i class='bi bi-{icon}' title='{label}'></i>"
                "</span>",
                class_=SafeString(class_),
                icon=current.icon,
                label=current.sort.label,
            )
    elif next_:
        return format_html(
            "<a href='?{query_string}' class='{class_}'>"
            "<i class='bi bi-{icon}' title='{label}'></i></a>",
            class_=SafeString(class_),
            query_string=next_.query_string,
            icon=next_.icon,
            label=next_.sort.label,
        )
    else:
        return SafeString()


class TableOrdering[M: Model]:
    """
    Instantiated ordering option for a table.

    This is a way to order a table, defining what is being ordered and in what
    way.
    """

    def __init__(
        self,
        orderings: "TableOrderings[M]",
        sort: Sort,
        what: str,
        how: str,
        label_what: str,
        query_string: str,
    ) -> None:
        """
        Ordering action bound to a table column.

        :param orderings: containing collection of table orderings
        :param sort: definition of how rows are ordered
        :param what: codename of what we are ordering
        :param how: codename of how we are ordering
        :param label_what: label describing what is being ordered
        :param query_string: URL query string to activate this ordering
        """
        self.orderings = orderings
        self.sort = sort
        self.what: str = what
        self.how: str = how
        self.label_what: str = label_what
        self.query_string: str = query_string
        #: Whether this is the current ordering
        #: (externally set at table instantiation)
        self.active: bool = False

    @override
    def __str__(self) -> str:
        """Return a stringified name."""
        return f"{self.what}.{self.how}"

    @override
    def __repr__(self) -> str:
        """Return a stringified name."""
        return f"""TableOrdering("{self.what}.{self.how}")"""

    @property
    def icon(self) -> str:
        """Icon representing the ordering."""
        return self.sort.icon

    @cached_property
    def table_menu_item(self) -> SafeString:
        """Render the ordering as a table-wide menu entry."""
        return format_html(
            "<a class='dropdown-item{active}' href='?{query_string}'"
            "{aria_current}><i class='bi bi-{icon}'></i> "
            "{label_what}: {label}</a>",
            label_what=self.label_what,
            label=self.sort.label,
            icon=self.icon,
            query_string=self.query_string,
            active=" active" if self.active else "",
            aria_current=(
                SafeString(" aria-current='true'") if self.active else ""
            ),
        )

    @cached_property
    def column_menu_item(self) -> SafeString:
        """Render the ordering as a column header menu entry."""
        return format_html(
            "<a class='dropdown-item{active}' href='?{query_string}'"
            "{aria_current}><i class='bi bi-{icon}'></i> {label}</a>",
            label=self.sort.label,
            icon=self.icon,
            query_string=self.query_string,
            active=" active" if self.active else "",
            aria_current=(
                SafeString(" aria-current='true'") if self.active else ""
            ),
        )


class TableOrderings[M: Model]:
    """All available table ordering options."""

    def __init__(self, table: "Table[M]") -> None:
        """Orderings for a table."""
        self.table = table
        #: Query string field name used to choose the table ordering
        self.field_name = self.table.prefix + "order"
        #: ID for this element in the DOM
        self.dom_id = self.table.dom_id + "orderings"
        #: Available sort options
        self.options: list[TableOrdering[M]] = []
        #: Active sort option
        self.current: TableOrdering[M]

    def finalize(self) -> None:
        """Finish setting up after the orderings definition is complete."""
        if not self.options:
            return
        # Resolve the default ordering
        default_ordering = self._find(self.table.default_order)
        if default_ordering is None:
            options = ", ".join(str(to) for to in self.options)
            raise ImproperlyConfigured(
                f"default_order {self.table.default_order!r}"
                " does not match available ordering options"
                f" ({options})"
            )

        # Resolve the requested ordering and pick the current one
        requested = self.table.request.GET.get(self.field_name)
        if requested is None:
            self.current = default_ordering
        else:
            self.current = self._find(requested) or default_ordering
        self.current.active = True

    def add(self, what: str, ordering: Ordering) -> list[TableOrdering[M]]:
        """Add an ordering definition."""
        added: list[TableOrdering[M]] = []
        for how, sort in ordering.options.items():
            table_ordering = TableOrdering(
                orderings=self,
                sort=sort,
                what=what,
                how=how,
                label_what=ordering.label or pretty_name(what),
                query_string=self.table.make_link(order=f"{what}.{how}"),
            )
            added.append(table_ordering)

        self.options += added
        return added

    def _find(self, value: str) -> TableOrdering[M] | None:
        """
        Find an ordering by what and how.

        :param what: ``what`` argument to match
        :param how: if None, returns the first entry for the given ``what``
        """
        what, how = self._parse_order(value)
        for to in self.options:
            if what != to.what:
                continue
            if how is None or how == to.how:
                return to
        return None

    def _parse_order(self, value: str) -> tuple[str, str | None]:
        """Resolve an order definition into a (what, how) tuple."""
        if value.startswith("-"):
            return value[1:], "desc"
        elif "." not in value:
            return value, None
        else:
            what, how = value.split(".", 1)
            return what, how

    @cached_property
    def what_menu_entries(self) -> list[SafeString]:
        """Return a list of menu entries for a what menu."""
        what_options: dict[str, TableOrdering[M]] = {}
        for o in self.options:
            if o.what in what_options:
                continue
            what_options[o.what] = o

        items: list[SafeString] = []
        for o in what_options.values():
            active = self.current.what == o.what
            items.append(
                format_html(
                    "<a class='dropdown-item{active}' href='?{query_string}'"
                    "{aria_current}>{label}</a>",
                    label=o.label_what,
                    query_string=o.query_string,
                    active=" active" if active else "",
                    aria_current=(
                        SafeString(" aria-current='true'") if active else ""
                    ),
                )
            )
        return items

    @cached_property
    def how_menu_entries(self) -> list[SafeString]:
        """Return a list of menu entries for a how menu for the current what."""
        hows = [o for o in self.options if o.what == self.current.what]
        items: list[SafeString] = []
        for o in hows:
            active = self.current == o
            items.append(
                format_html(
                    "<a class='dropdown-item{active}' href='?{query_string}'"
                    "{aria_current}>{label}</a>",
                    label=o.sort.label,
                    query_string=o.query_string,
                    active=" active" if active else "",
                    aria_current=(
                        SafeString(" aria-current='true'") if active else ""
                    ),
                )
            )
        return items

    @cached_property
    def how_sort_icon(self) -> SafeString:
        """Return a sort icon for the current hows."""
        hows = [o for o in self.options if o.what == self.current.what]
        return render_sort_icon(hows, for_table=True)

    def order_queryset(self, queryset: QuerySet[M, M]) -> QuerySet[M, M]:
        """Order a queryset."""
        return queryset.order_by(*self.current.sort.order_by)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used by debusine server-side tasks."""

import datetime as dt
import re
from typing import Annotated, Any, Self, override

import pydantic

from debusine.artifacts.models import WorkRequestResults
from debusine.db.models.workspaces import workspace_name_regex
from debusine.tasks.models import (
    BaseDynamicTaskData,
    BaseTaskData,
    BaseTaskDataModel,
    LookupMultiple,
    LookupSingle,
)


class ServerNoopData(BaseTaskData):
    """In-memory task data for the ServerNoop task."""

    exception: bool = False
    result: WorkRequestResults = WorkRequestResults.FAILURE


class APTMirrorFilter(BaseTaskDataModel):
    """A filter for APT repository mirroring."""

    binary_name: re.Pattern[str] | None = None
    priority: str | None = None
    section: str | None = None
    source_name: re.Pattern[str] | None = None

    ignore_unauthorized: bool = False


class APTMirrorData(BaseTaskData):
    """In-memory task data for the APTMirror task."""

    collection: LookupSingle
    url: pydantic.AnyUrl
    suite: str
    components: list[str] | None = None
    architectures: list[str] | None = None
    signing_key: str | None = None
    filters: list[APTMirrorFilter] | None = None
    authentication: str | None = None

    @pydantic.model_validator(mode="after")
    def check_suite_components_consistency(self) -> Self:
        """Components are only allowed/required for non-flat repositories."""
        if self.suite.endswith("/"):
            if self.components is not None:
                raise ValueError(
                    'Flat repositories (where suite ends with "/") must not '
                    'have components'
                )
        else:
            if self.components is None:
                raise ValueError(
                    'Non-flat repositories (where suite does not end with '
                    '"/") must have components'
                )
        return self


class PackageUploadInput(BaseTaskDataModel):
    """Input for the PackageUpload task."""

    upload: LookupSingle


class PackageUploadTarget(pydantic.AnyUrl):
    """
    Target URL for the PackageUpload task.

    While this may contain a user and password, note that these are not
    currently secret.
    """

    allowed_schemes = {"ftp", "sftp"}


class PackageUploadData(BaseTaskData):
    """In-memory task data for the PackageUpload task."""

    input: PackageUploadInput
    target: PackageUploadTarget
    delayed_days: int | None = None


class PackageUploadDynamicData(BaseDynamicTaskData):
    """Dynamic data for the PackageUpload task."""

    input_upload_id: int

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [self.input_upload_id]


class CopyCollectionItemsCopies(BaseTaskDataModel):
    """Specification of a set of copies for the CopyCollectionItems task."""

    source_items: LookupMultiple
    target_collection: LookupSingle
    unembargo: bool = False
    replace: bool = False
    name_template: str | None = None
    variables: dict[str, Any] | None = None


class CopyCollectionItemsData(BaseTaskData):
    """In-memory task data for the CopyCollectionItems task."""

    copies: Annotated[
        list[CopyCollectionItemsCopies], pydantic.Field(min_length=1)
    ]


class CreateChildWorkspaceData(BaseTaskData):
    """In-memory task data for CreateChildWorkspace."""

    prefix: Annotated[str, pydantic.Field(pattern=workspace_name_regex)]
    suffix: Annotated[str, pydantic.Field(pattern=workspace_name_regex)]
    public: bool = True
    owner_group: str | None = None
    workflow_template_names: list[str] = []
    expiration_delay: int | None = 60


class GenerateSuiteIndexesData(BaseTaskData):
    """In-memory task data for GenerateSuiteIndexes."""

    suite_collection: LookupSingle
    generate_at: dt.datetime


class GenerateSuiteIndexesDynamicData(BaseDynamicTaskData):
    """Dynamic data for the GenerateSuiteIndexes task."""

    suite_collection_id: int



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Middleware for maintaining debusine.db.context."""

from collections.abc import Awaitable, Callable
from inspect import iscoroutinefunction
from typing import TYPE_CHECKING, override

import django.http
from asgiref.typing import ASGIReceiveCallable, ASGISendCallable
from django.utils.deprecation import MiddlewareMixin

if TYPE_CHECKING:
    from channels.consumer import _ChannelScope


class ContextMiddleware(MiddlewareMixin):
    """Run get_response in a private contextvar context."""

    def process_request(
        self,
        request: django.http.HttpRequest,  # noqa: ARG002, U100
    ) -> None:
        """Process request in middleware."""
        from debusine.db.context import context

        # Make application context request-local
        context.reset()

    @override
    async def __acall__(
        self, request: django.http.HttpRequest
    ) -> django.http.HttpResponseBase:
        """Asynchronous middleware entry point."""
        # MiddlewareMixin provides __acall__, but its implementation calls
        # process_request within sync_to_async.  In this case, we must call
        # it in the async context: sync_to_async copies context variables
        # that have been set in the async context into and out of its
        # thread, but it doesn't copy out context variables that haven't
        # been set at all in the async context.
        self.process_request(request)
        assert iscoroutinefunction(self.get_response)
        response = await self.get_response(request)
        assert isinstance(response, django.http.HttpResponseBase)
        return response


class ContextMiddlewareChannels:
    """Run the app in a private contextvar context."""

    def __init__(
        self,
        app: Callable[
            ["_ChannelScope", ASGIReceiveCallable, ASGISendCallable],
            Awaitable[None],
        ],
    ) -> None:
        """Middleware API entry point."""
        self.app = app

    async def __call__(
        self,
        scope: "_ChannelScope",
        receive: ASGIReceiveCallable,
        send: ASGISendCallable,
    ) -> None:
        """Middleware entry point."""
        from debusine.db.context import context

        # Make application context request-local
        context.reset()
        return await self.app(scope, receive, send)



# Generated by Django 5.2.13 on 2026-05-06 16:14

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('db', '0057_filestorejournal'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='user',
            constraint=models.CheckConstraint(
                check=models.Q(('username__regex', '^\\d+$'), _negated=True),
                name='db_user_non_integer_username',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for work requests."""

from collections import defaultdict
from functools import cached_property
from typing import Any, override

from django.utils.formats import date_format
from django.utils.html import format_html
from django.utils.timesince import timesince

from debusine.artifacts.models import TaskTypes
from debusine.db.models import WorkRequest, WorkflowTemplate
from debusine.db.models.work_requests import InternalTaskError
from debusine.db.tasks import DBTask
from debusine.tasks import TaskConfigError
from debusine.web.helps import DOCS_BASE_URL
from debusine.web.icons import Icons
from debusine.web.views import ui_shortcuts
from debusine.web.views.places import ResourcePlace
from debusine.web.views.sidebar import (
    SidebarAction,
    SidebarElement,
    SidebarItem,
)
from debusine.web.views.ui.base import UI


class WorkflowTemplateUI(UI[WorkflowTemplate]):
    """UI helpers for WorkflowTemplate instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this workflow template."""
        return ResourcePlace(
            title=f"Workflow template {self.instance.name}",
            url=self.instance.get_absolute_url(),
            breadcrumb=self.instance.name,
            parent=self.instance.workspace.ui(self.request).place,
        )

    def get_help_url(self) -> str:
        """Return a URL to the appropriate workflow documentation."""
        # We could avoid hardcoding this URL structure by using Sphinx to
        # fetch and parse the intersphinx inventory.  However, Sphinx seems
        # like too heavy a dependency to pull in just for this.
        return (
            f"{DOCS_BASE_URL}/reference/workflows/specs/"
            f"{self.instance.task_name.replace('_', '-')}.html"
        )


class WorkRequestUI(UI[WorkRequest]):
    """UI helpers for WorkRequest instances."""

    @cached_property
    def task(self) -> DBTask[Any] | None:
        """Return the task for this work request, if it can be instantiated."""
        try:
            return self.instance.get_task()
        except (TaskConfigError, InternalTaskError):
            return None

    @cached_property
    def label(self) -> str:
        """Get the label for this work request."""
        if self.task is not None:
            return self.task.get_label()
        elif (
            self.instance.is_workflow
            and self.instance.workflow_data.workflow_template_name
        ):
            return self.instance.workflow_data.workflow_template_name
        else:
            return self.instance.task_name

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this work request."""
        is_workflow_root = self.instance == self.instance.workflow_root
        place = ResourcePlace(
            title=f"#{self.instance.id}: {self.label}",
            description=f"Work request #{self.instance.id}: {self.label}",
            breadcrumb=(
                f"#{self.instance.id}: {self.label}"
                if is_workflow_root
                else f"#{self.instance.id}"
            ),
            url=self.instance.get_absolute_url(),
            icon=Icons.WORK_REQUEST,
        )
        if self.instance.parent is not None:
            place.parent = self.instance.parent.ui(self.request).place
        elif self.instance.task_type == TaskTypes.WORKFLOW:
            place.parent = self.instance.workspace.ui(
                self.request
            ).place_workflow_list
        else:
            place.parent = self.instance.workspace.ui(self.request).place
        return place

    @cached_property
    def workflow_children(self) -> dict[int, list[WorkRequest]]:
        """Mapping from work request IDs to their children in this workflow."""
        children: defaultdict[int, list[WorkRequest]] = defaultdict(list)
        for wr in (
            WorkRequest.objects.filter(id=self.instance.id)
            .workflow_flattened(include_workflows=True)
            .filter(parent__isnull=False)
            .visible_in_workflow()
            .order_by("id")
            .select_related(
                "workspace__scope",
                "aborted_by",
            )
        ):
            assert wr.parent_id is not None
            children[wr.parent_id].append(wr)
        return children

    @override
    def build_ui_shortcuts(self) -> list[ui_shortcuts.UIShortcut]:
        result = super().build_ui_shortcuts()
        if self.request.user.is_authenticated:
            if self.instance.verify_retry():
                result.append(
                    ui_shortcuts.create_work_request_retry(self.instance)
                )
            if self.instance.verify_abort():
                result.append(
                    ui_shortcuts.create_work_request_abort(self.instance)
                )
        return result

    def sidebar_work_request(self) -> SidebarItem:
        """Create sidebar item with work request information."""
        kwargs = {
            "icon": Icons.WORK_REQUEST,
            "label": "Work request",
        }

        kwargs["detail"] = (
            f"{self.instance.task_type} {self.instance.task_name} task"
        )
        kwargs["value"] = self.label

        return SidebarAction(
            url=self.instance.get_absolute_url(),
            **kwargs,
        )

    def sidebar_work_request_status(self) -> SidebarItem:
        """Create information about a work request status and result."""
        kwargs = {
            "icon": Icons.WORK_REQUEST_STATUS,
            "label": "Status and result",
        }

        match self.instance.status:
            case WorkRequest.Statuses.PENDING:
                status_bg = "secondary"
            case WorkRequest.Statuses.RUNNING:
                status_bg = "info"
            case WorkRequest.Statuses.COMPLETED:
                status_bg = "primary"
            case WorkRequest.Statuses.ABORTED | WorkRequest.Statuses.BLOCKED:
                status_bg = "dark"
            case _ as unreachable:  # noqa: F841
                status_bg = "info"

        status = format_html(
            '<span class="badge text-bg-{bg}">{status}</span>',
            bg=status_bg,
            status=self.instance.status.capitalize(),
        )

        result_bg: str | None
        match self.instance.result:
            case WorkRequest.Results.NONE:
                result_bg = None
            case WorkRequest.Results.SUCCESS:
                result_bg = "success"
            case WorkRequest.Results.FAILURE:
                result_bg = "warning"
            case WorkRequest.Results.ERROR:
                result_bg = "danger"
            case WorkRequest.Results.SKIPPED:
                result_bg = "secondary"
            case _ as unreachable:  # noqa: F841
                result_bg = "info"

        if result_bg is None:
            kwargs["value"] = status
        else:
            kwargs["value"] = status + format_html(
                ' <span class="badge text-bg-{bg}">{result}</span>',
                bg=result_bg,
                result=self.instance.result.capitalize(),
            )

        return SidebarElement(**kwargs)

    def sidebar_work_request_superseded(self) -> SidebarItem:
        """Create a link to the work request that supersedes this one."""
        kwargs = {
            "icon": Icons.WORK_REQUEST_SUPERSEDED,
            "label": "Superseded by",
        }
        if (superseded := getattr(self.instance, "superseded", None)) is None:
            return SidebarElement(**kwargs)
        return SidebarAction(
            url=superseded.get_absolute_url(),
            value=str(superseded),
            **kwargs,
        )

    def sidebar_work_request_supersedes(self) -> SidebarItem:
        """Create a link to the work request superseded by this one."""
        kwargs = {
            "icon": Icons.WORK_REQUEST_SUPERSEDES,
            "label": "Supersedes",
        }
        if (supersedes := getattr(self.instance, "supersedes", None)) is None:
            return SidebarElement(**kwargs)

        return SidebarAction(
            url=supersedes.get_absolute_url(),
            value=str(supersedes),
            **kwargs,
        )

    def sidebar_workflow(self) -> SidebarItem:
        """Create a link to the associated workflow."""
        workflow = self.instance.workflow_root
        if workflow is None:
            return SidebarElement(
                icon=Icons.WORKFLOW,
                label="Workflow",
            )

        return SidebarAction(
            icon="diagram-3",
            label="Workflow",
            url=workflow.get_absolute_url(),
            value=str(workflow),
        )

    def sidebar_work_request_started_at(self) -> SidebarItem:
        """Show the queue time of the work request before being started."""
        kwargs = {
            "icon": Icons.STARTED_AT,
            "label": "Time from creation to start",
        }
        if not self.instance.started_at:
            return SidebarElement(**kwargs)

        return SidebarElement(
            detail=date_format(self.instance.started_at, "DATETIME_FORMAT"),
            value=timesince(self.instance.created_at, self.instance.started_at),
            **kwargs,
        )

    def sidebar_work_request_duration(self) -> SidebarItem:
        """Show the duration of the work request run."""
        kwargs = {
            "icon": Icons.DURATION,
            "label": "Duration",
        }

        if (
            self.instance.started_at
            and self.instance.status == WorkRequest.Statuses.RUNNING
        ):
            return SidebarElement(
                value=timesince(self.instance.started_at),
                **kwargs,
            )

        if not self.instance.started_at or not self.instance.completed_at:
            return SidebarElement(**kwargs)

        return SidebarElement(
            value=timesince(
                self.instance.started_at, self.instance.completed_at
            ),
            detail=date_format(self.instance.completed_at, "DATETIME_FORMAT"),
            **kwargs,
        )



"""
Override path-related settings to use system wide paths.

The paths indicated in this file are those setup by the official Debian
package.
"""

__all__ = [
    "DEBUSINE_CACHE_DIRECTORY",
    "DEBUSINE_DATA_PATH",
    "DEBUSINE_LOG_DIRECTORY",
    "DEBUSINE_TEMPLATE_DIRECTORY",
]

DEBUSINE_DATA_PATH = "/var/lib/debusine/server"
DEBUSINE_CACHE_DIRECTORY = "/var/cache/debusine/server"
DEBUSINE_LOG_DIRECTORY = "/var/log/debusine/server"
DEBUSINE_TEMPLATE_DIRECTORY = "/etc/debusine/server/templates"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Make signed source workflow."""

from functools import cached_property
from typing import override

from debusine.artifacts.models import ArtifactCategory
from debusine.assets import KeyPurpose
from debusine.client.models import LookupChildType
from debusine.db.models import WorkRequest
from debusine.server.collections.lookup import artifact_ids
from debusine.server.workflows import Workflow, workflow_utils
from debusine.server.workflows.models import (
    MakeSignedSourceWorkflowData,
    MakeSignedSourceWorkflowDynamicData,
    SbuildWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.signing.tasks.models import SignData
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    AssembleSignedSourceData,
    BackendType,
    ExtractForSigningData,
    ExtractForSigningInput,
    InputArtifact,
    InputArtifactMultiple,
    LookupDict,
    LookupMultiple,
    LookupSingle,
    SbuildInput,
)


class MakeSignedSourceWorkflow(
    Workflow[MakeSignedSourceWorkflowData, MakeSignedSourceWorkflowDynamicData]
):
    """Make signed sources."""

    TASK_NAME = "make_signed_source"

    @override
    def compute_dynamic_data(self) -> MakeSignedSourceWorkflowDynamicData:
        """
        Compute dynamic data.

        :subject: source package names separated by spaces
        :parameter_summary: source package names separated by spaces followed
          by "for ``vendor``:``codename``"
        :binary_artifacts_ids: ids of the binary_artifacts
        :signing_template_artifacts_ids: ids of the signing_template_artifacts
        """
        binaries = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )

        source_package_names = " ".join(
            workflow_utils.get_source_package_names(
                binaries,
                configuration_key="binary_artifacts",
                artifact_expected_categories=(
                    ArtifactCategory.BINARY_PACKAGE,
                    ArtifactCategory.UPLOAD,
                ),
            )
        )

        signing_templates = self.work_request.lookup_multiple(
            self.data.signing_template_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )

        return MakeSignedSourceWorkflowDynamicData(
            subject=source_package_names,
            parameter_summary=f"{source_package_names} for "
            f"{self.data.vendor}:{self.data.codename}",
            binary_artifacts_ids=artifact_ids(binaries),
            signing_template_artifacts_ids=artifact_ids(signing_templates),
        )

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return input artifacts."""
        return [
            InputArtifactMultiple.from_ids(
                lookup=self.data.binary_artifacts,
                label="binary_artifacts",
                artifact_ids=(
                    self.dynamic_data.binary_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            ),
            InputArtifactMultiple.from_ids(
                lookup=self.data.signing_template_artifacts,
                label="signing_template_artifacts",
                artifact_ids=(
                    self.dynamic_data.signing_template_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            ),
        ]

    @cached_property
    def architectures(self) -> set[str]:
        """Architectures to run on."""
        return (
            workflow_utils.get_lookup_architectures(
                self, self.data.signing_template_artifacts
            )
            .intersection(
                workflow_utils.get_lookup_architectures(
                    self, self.data.binary_artifacts
                )
            )
            .intersection(self.data.architectures)
        )

    @override
    def populate(self) -> None:
        """Create work requests and sub-workflows."""
        workflow_root = self.work_request.workflow_root
        environment = f"{self.data.vendor}/match:codename={self.data.codename}"

        signing_templates = self.work_request.lookup_multiple(
            self.data.signing_template_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )

        for arch in sorted(self.architectures):
            signing_template_lookups_by_name = {
                workflow_utils.lookup_result_binary_package_name(
                    result
                ): result.reconstruct(workflow_root=workflow_root)
                for result in signing_templates
                if workflow_utils.lookup_result_architectures(result) == {arch}
            }

            binary_artifacts_lookups = (
                workflow_utils.filter_artifact_lookup_by_arch(
                    self,
                    self.data.binary_artifacts,
                    {arch, "all"},
                )
            )

            for signing_template_name, signing_template_lookup in sorted(
                signing_template_lookups_by_name.items()
            ):
                # filter_artifact_lookup_by_arch always returns a parsed list of
                # single lookups.
                assert not isinstance(signing_template_lookup, LookupDict)

                extract_for_signing_wr, extracted_lookup = (
                    self._populate_extract_for_signing(
                        template_artifact=signing_template_lookup,
                        binary_artifacts=binary_artifacts_lookups,
                        environment=environment,
                        architecture=arch,
                        template_name=signing_template_name,
                    )
                )

                sign_wr, signed = self._populate_sign(
                    purpose=self.data.purpose,
                    unsigned=extracted_lookup,
                    key=self.data.key,
                    extract_for_signing_wr=extract_for_signing_wr,
                    architecture=arch,
                    template_name=signing_template_name,
                )

                assembled = self._populate_assemble_signed_source(
                    environment=environment,
                    template=signing_template_lookup,
                    signed=signed,
                    sign_wr=sign_wr,
                    architecture=arch,
                    template_name=signing_template_name,
                )

                target_distribution = f"{self.data.vendor}:{self.data.codename}"

                sbuild_architectures = sorted(self.architectures | {"all"})
                self._populate_sbuild_workflow(
                    assembled=assembled,
                    extra_binary_artifacts=binary_artifacts_lookups,
                    target_distribution=target_distribution,
                    backend=self.data.sbuild_backend,
                    architectures=sbuild_architectures,
                    architecture=arch,
                    template_name=signing_template_name,
                )

    def _populate_extract_for_signing(
        self,
        *,
        template_artifact: LookupSingle,
        binary_artifacts: LookupMultiple,
        environment: str,
        architecture: str,
        template_name: str,
    ) -> tuple[WorkRequest, LookupMultiple]:
        """
        Create work request for ExtractForSigning.

        :returns: A tuple of the ExtractForSigning work request and a lookup
          of the extracted artifacts in the collection.
        """
        wr, _ = self.work_request_ensure_child_worker(
            task_name="extractforsigning",
            task_data=ExtractForSigningData(
                environment=environment,
                input=ExtractForSigningInput(
                    template_artifact=template_artifact,
                    binary_artifacts=binary_artifacts,
                ),
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=(
                    f"Extract for signing ({template_name}/{architecture})"
                ),
                step=f"extract-for-signing-{architecture}-{template_name}",
            ),
        )
        self.requires_artifact(wr, template_artifact)
        self.requires_artifact(wr, binary_artifacts)

        wr.add_event_reaction(
            "on_success",
            ActionUpdateCollectionWithArtifacts(
                collection="internal@collections",
                name_template=(
                    "extracted-for-signing-{architecture}-{binary_package_name}"
                ),
                variables={
                    # Extract the binary package name from the artifact
                    # data, since the workflow doesn't know it up-front.
                    "$binary_package_name": "binary_package_name",
                    "architecture": architecture,
                },
                artifact_filters={"category": ArtifactCategory.SIGNING_INPUT},
            ),
        )
        lookup = LookupMultiple.model_validate(
            {
                "collection": "internal@collections",
                "child_type": LookupChildType.ARTIFACT,
                "category": ArtifactCategory.SIGNING_INPUT,
                "name__startswith": "extracted-for-signing-",
                "data__architecture": architecture,
            }
        )
        return wr, lookup

    def _populate_sign(
        self,
        *,
        purpose: KeyPurpose,
        unsigned: LookupMultiple,
        key: str,
        extract_for_signing_wr: WorkRequest,
        architecture: str,
        template_name: str,
    ) -> tuple[WorkRequest, LookupMultiple]:
        """
        Create work request for Sign.

        :returns: A tuple of the Sign work request and a lookup of the
          signed artifacts in the collection.
        """
        wr, _ = self.work_request_ensure_child_signing(
            task_name="sign",
            task_data=SignData(purpose=purpose, unsigned=unsigned, key=key),
            workflow_data=WorkRequestWorkflowData(
                display_name=f"Sign ({template_name}/{architecture})",
                step=f"sign-{architecture}-{template_name}",
            ),
        )

        wr.add_dependency(extract_for_signing_wr)

        wr.add_event_reaction(
            "on_success",
            ActionUpdateCollectionWithArtifacts(
                collection="internal@collections",
                name_template=(
                    "signed-{architecture}-{template_name}-"
                    "{binary_package_name}"
                ),
                variables={
                    # Extract the binary package name from the artifact
                    # data, since the workflow doesn't know it up-front.
                    "$binary_package_name": "binary_package_name",
                    "architecture": architecture,
                    "template_name": template_name,
                },
                artifact_filters={"category": ArtifactCategory.SIGNING_OUTPUT},
            ),
        )
        lookup = LookupMultiple.model_validate(
            {
                "collection": "internal@collections",
                "child_type": LookupChildType.ARTIFACT,
                "category": ArtifactCategory.SIGNING_OUTPUT,
                "name__startswith": f"signed-{architecture}-{template_name}-",
                "data__architecture": architecture,
            }
        )
        return wr, lookup

    def _populate_assemble_signed_source(
        self,
        *,
        environment: str,
        template: LookupSingle,
        signed: LookupMultiple,
        sign_wr: WorkRequest,
        architecture: str,
        template_name: str,
    ) -> LookupSingle:
        """
        Create work request for assembling the signed source.

        :returns: Lookup of the assembled signed artifact
        """
        wr, _ = self.work_request_ensure_child_worker(
            task_name="assemblesignedsource",
            task_data=AssembleSignedSourceData(
                environment=environment, template=template, signed=signed
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=(
                    f"Assemble signed source ({template_name}/{architecture})"
                ),
                step=f"assemble-signed-source-{architecture}-{template_name}",
            ),
        )

        wr.add_dependency(sign_wr)

        artifact_name = f"signed-source-{architecture}-{template_name}"
        self.provides_artifact(wr, ArtifactCategory.UPLOAD, artifact_name)
        return f"internal@collections/name:{artifact_name}"

    def _populate_sbuild_workflow(
        self,
        assembled: LookupSingle,
        extra_binary_artifacts: LookupMultiple,
        target_distribution: str,
        backend: BackendType,
        architectures: list[str],
        architecture: str,
        template_name: str,
    ) -> None:
        """Populate SbuildWorkflow."""
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="sbuild",
            task_data=SbuildWorkflowData(
                prefix=f"signed-source-{architecture}-{template_name}|",
                input=SbuildInput(
                    source_artifact=assembled,
                    extra_binary_artifacts=extra_binary_artifacts,
                ),
                target_distribution=target_distribution,
                backend=backend,
                architectures=architectures,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=f"Sbuild ({template_name}/{architecture})",
                step=f"sbuild-{architecture}-{template_name}",
            ),
        )
        self.requires_artifact(wr, assembled)
        self.orchestrate_child(wr)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-signing command to run a signing worker."""

import asyncio
import logging
from typing import Any, override

from django.core.management import CommandParser

from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.tasks.models import WorkerType
from debusine.worker import Worker


class Command(DebusineBaseCommand):
    """Command to run a signing worker."""

    help = "Run a signing worker"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        """Add CLI arguments for the signing_worker command."""
        parser.add_argument(
            "--log-level",
            help=(
                f"Minimum log level. Overrides log-level (in [General] "
                f"section) from config.ini. If not specified anywhere logs "
                f"{logging.getLevelName(Worker.DEFAULT_LOG_LEVEL)} and above."
            ),
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        )

    @override
    def handle(self, *args: Any, **options: Any) -> None:
        """Run a signing worker."""
        worker = Worker(
            log_level=options["log_level"],
            worker_type=WorkerType.SIGNING,
        )
        asyncio.run(worker.main())

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli class."""

import contextlib
import io
import signal
from unittest import mock

from debusine.client.cli import Cli
from debusine.client.config import ConfigHandler
from debusine.client.tests.base import BaseCliTests


class CliTests(BaseCliTests):
    """Tests for the debusine command line interface generic functionality."""

    def test_client_invalid_subcommand(self) -> None:
        cli = self.create_cli(["does-not-exist"], create_config=False)
        stderr = io.StringIO()

        with contextlib.redirect_stderr(stderr), self.assertRaisesSystemExit(2):
            # Argparse prints help and exits with exit_code=2
            cli.execute()

        self.assertRegex(
            stderr.getvalue(),
            r"debusine: error: argument sub-command: invalid choice:"
            r" 'does-not-exist' \(choose from",
        )

    def test_client_subcommand_not_handled(self) -> None:
        cli = self.create_cli(["artifact", "show", "1"], create_config=False)
        with (
            mock.patch(
                "debusine.client.commands.base.Command.create",
                return_value=None,
            ),
            self.assertRaisesRegex(
                AssertionError, "Unexpected sub-command name: artifact"
            ),
        ):
            cli.execute()

    def test_client_without_parameters(self) -> None:
        """
        Executing the client without any parameter returns an error.

        At least one subcommands is required. argparse prints help
        and exit.
        """
        cli = self.create_cli([], create_config=False)
        stderr = io.StringIO()

        with contextlib.redirect_stderr(stderr), self.assertRaisesSystemExit(2):
            # Argparse prints help and exits with exit_code=2
            cli.execute()

        self.assertGreater(len(stderr.getvalue()), 80)

    def test_client_help_include_default_setting(self) -> None:
        """Cli.execute() help include ConfigHandler.DEFAULT_CONFIG_FILE_PATH."""
        stdout = io.StringIO()

        cli = self.create_cli(['--help'], create_config=False)

        with contextlib.redirect_stdout(stdout), self.assertRaisesSystemExit(0):
            # Argparse prints help and exits with exit_code=0
            cli.execute()

        # argparse might add \n and spaces (for indentation) in the
        # output to align the text. In this case
        # ConfigHandler.DEFAULT_CONFIG_FILE_PATH could not be found
        output = stdout.getvalue().replace('\n', '').replace(' ', '')

        self.assertIn(
            str(ConfigHandler.DEFAULT_CONFIG_FILE_PATH).replace(' ', ''), output
        )

    def test_client_version(self) -> None:
        stdout = io.StringIO()
        cli = self.create_cli(["--version"], create_config=False)

        with (
            contextlib.redirect_stdout(stdout),
            self.assertRaisesSystemExit(0),
            mock.patch(
                "debusine.client.cli.debusine_version", return_value="0.1.0"
            ),
        ):
            cli.execute()

        self.assertEqual("debusine 0.1.0\n", stdout.getvalue())

    def test_client_version_uninstalled(self) -> None:
        stdout = io.StringIO()
        cli = self.create_cli(["--version"], create_config=False)

        with (
            contextlib.redirect_stdout(stdout),
            self.assertRaisesSystemExit(0),
            mock.patch(
                "debusine.client.cli.debusine_version", return_value=None
            ),
        ):
            cli.execute()

        self.assertEqual("debusine uninstalled-version\n", stdout.getvalue())

    def test_exit_raise_system_exit_0(self) -> None:
        """Cli._exit raise SystemExit(0)."""
        with self.assertRaisesSystemExit(0):
            Cli._exit(signal.SIGINT, None)

    def test_signal_handlers_set_on_exec(self) -> None:
        """Test that the signal handlers are set on exec()."""
        cli = self.create_cli(["show-work-request", "1"])

        # Cli.__init__() is not changing the signal handlers for SIG{INT,TERM}
        self.assertEqual(
            signal.getsignal(signal.SIGINT), self.default_sigint_handler
        )
        self.assertEqual(
            signal.getsignal(signal.SIGTERM), self.default_sigterm_handler
        )

        with mock.patch(
            "debusine.client.commands.work_requests.Show.run",
            autospec=True,
        ):
            cli.execute()

        # cli.execute() changed the signal handlers for SIG{INT,TERM}
        self.assertEqual(signal.getsignal(signal.SIGINT), cli._exit)
        self.assertEqual(signal.getsignal(signal.SIGTERM), cli._exit)
